from typing import List, Dict, Optional

from pydantic import BaseModel


class TestInputDTO(BaseModel):
    name: str
    value: str
    type: str
    params: List["TestInputDTO"] = list()
    is_alias: bool = False


class SuiteTestDTO(BaseModel):
    testUuid: str
    functionInputs: Dict[str, TestInputDTO]


class TestSuiteDTO(BaseModel):
    name: Optional[str]
    project_key: str
    tests: List[SuiteTestDTO]



from grpc import StatusCode


class CodedError(Exception):
    code: StatusCode = StatusCode.INTERNAL

    def __init__(self, code: StatusCode, *args: object) -> None:
        super().__init__(*args)
        self.code = code


class IllegalArgumentError(CodedError):
    def __init__(self, *args: object) -> None:
        super().__init__(StatusCode.INVALID_ARGUMENT, *args)



import datetime
from time import perf_counter
import pandas as pd
from typing import Sequence
from abc import abstractmethod

from ...slicing.slice_finder import SliceFinder

from ..registry import Detector

from ...models.base import BaseModel
from ...datasets.base import Dataset
from ...ml_worker.testing.registry.slicing_function import SlicingFunction
from ..logger import logger
from ..issues import Issue


class LossBasedDetector(Detector):
    MIN_DATASET_LENGTH = 100
    MAX_DATASET_SIZE = 10_000_000
    LOSS_COLUMN_NAME = "__gsk__loss"

    _needs_target = True

    def run(self, model: BaseModel, dataset: Dataset):
        if self._needs_target and dataset.target is None:
            logger.info(f"{self.__class__.__name__}: Skipping detection because the dataset has no target column.")
            return []

        logger.info(f"{self.__class__.__name__}: Running")

        # Check if we have enough data to run the scan
        if len(dataset) < self.MIN_DATASET_LENGTH:
            logger.warning(
                f"{self.__class__.__name__}: Skipping scan because the dataset is too small"
                f" (< {self.MIN_DATASET_LENGTH} samples)."
            )
            return []

        # If the dataset is very large, limit to a subsample
        max_data_size = self.MAX_DATASET_SIZE // len(model.meta.feature_names or dataset.columns)
        if len(dataset) > max_data_size:
            logger.info(f"{self.__class__.__name__}: Limiting dataset size to {max_data_size} samples.")
            dataset = dataset.slice(lambda df: df.sample(max_data_size, random_state=42), row_level=False)

        # Calculate loss
        logger.info(f"{self.__class__.__name__}: Calculating loss")
        start = perf_counter()
        meta = self._calculate_loss(model, dataset)
        elapsed = perf_counter() - start
        logger.info(f"{self.__class__.__name__}: Loss calculated (took {datetime.timedelta(seconds=elapsed)})")

        # Find slices
        logger.info(f"{self.__class__.__name__}: Finding data slices")
        start = perf_counter()
        slices = self._find_slices(model, dataset, meta)
        elapsed = perf_counter() - start
        logger.info(
            f"{self.__class__.__name__}: {len(slices)} slices found (took {datetime.timedelta(seconds=elapsed)})"
        )

        # Create issues from the slices
        logger.info(f"{self.__class__.__name__}: Analyzing issues")
        start = perf_counter()
        issues = self._find_issues(slices, model, dataset, meta)
        elapsed = perf_counter() - start
        logger.info(
            f"{self.__class__.__name__}: {len(issues)} issues found (took {datetime.timedelta(seconds=elapsed)})"
        )

        return issues

    @property
    def _numerical_slicer_method(self):
        return "tree"

    def _find_slices(self, model: BaseModel, dataset: Dataset, meta: pd.DataFrame):
        features = model.meta.feature_names or dataset.columns.drop(dataset.target, errors="ignore")

        df_with_meta = dataset.df.join(meta, how="right")

        column_types = dataset.column_types.copy()
        column_types[self.LOSS_COLUMN_NAME] = "numeric"
        dataset_with_meta = Dataset(
            df_with_meta,
            target=dataset.target,
            column_types=column_types,
            validation=False,
        )

        # For performance
        dataset_with_meta.load_metadata_from_instance(dataset.column_meta)

        # Find slices
        sf = SliceFinder(numerical_slicer=self._numerical_slicer_method)
        sliced = sf.run(dataset_with_meta, features, target=self.LOSS_COLUMN_NAME)
        slices = sum(sliced.values(), start=[])

        # Keep only slices of size at least 5% of the dataset or 20 samples (whatever is larger)
        slices = [s for s in slices if max(0.05 * len(dataset), 20) <= len(dataset_with_meta.slice(s))]

        return slices

    @abstractmethod
    def _calculate_loss(self, model: BaseModel, dataset: Dataset) -> pd.DataFrame:
        ...

    @abstractmethod
    def _find_issues(
        self,
        slices: Sequence[SlicingFunction],
        model: BaseModel,
        dataset: Dataset,
        meta: pd.DataFrame,
    ) -> Sequence[Issue]:
        ...



import logging
from pathlib import Path
from typing import Union, Optional, Iterable, Any, Callable

import pandas as pd
import yaml
from scipy import special

from giskard.core.core import ModelType
from giskard.core.validation import configured_validate_arguments
from giskard.models.base import WrapperModel

try:
    import torch
except ImportError as e:
    raise ImportError("Please install it via 'pip install torch'") from e

try:
    from transformers import pipeline, pipelines
except ImportError as e:
    raise ImportError("Please install it via 'pip install transformers'") from e

logger = logging.getLogger(__name__)


class HuggingFaceModel(WrapperModel):
    """
    A subclass of the WrapperModel class for using HuggingFace transformers.

    The HuggingFaceModel class is a wrapper around the HuggingFace transformers library.

    Inherits all attributes and methods from the WrapperModel class.

    Attributes:
        huggingface_module (Type): The type of the HuggingFace module used by the model.
        pipeline_task (str, optional): The task performed by the HuggingFace pipeline, if applicable.
    """

    @configured_validate_arguments
    def __init__(
        self,
        model,
        model_type: ModelType,
        name: Optional[str] = None,
        data_preprocessing_function: Callable[[pd.DataFrame], Any] = None,
        model_postprocessing_function: Callable[[Any], Any] = None,
        feature_names: Optional[Iterable] = None,
        classification_threshold: Optional[float] = 0.5,
        classification_labels: Optional[Iterable] = None,
        **kwargs,
    ) -> None:
        """
        Initializes an instance of a HuggingFaceModel with the provided arguments and sets necessary attributes.

        Arguments:
            model: The model to be wrapped.
            model_type (ModelType): The type of the model, either regression or classification.
            name (str, optional): The name of the model.
            data_preprocessing_function (Callable[[pd.DataFrame], Any], optional): A function to preprocess the input data.
            model_postprocessing_function (Callable[[Any], Any], optional): A function to postprocess the model output.
            feature_names (Iterable, optional): The names of the model features.
            classification_threshold (float, optional): The classification probability threshold for binary classification models.
            classification_labels (Iterable, optional): The labels for classification models.

        Notes:
            huggingface_module (Type): The type of the HuggingFace module used by the model.
            pipeline_task (str, optional): The task performed by the HuggingFace pipeline, if applicable.
        """

        super().__init__(
            model=model,
            model_type=model_type,
            name=name,
            data_preprocessing_function=data_preprocessing_function,
            model_postprocessing_function=model_postprocessing_function,
            feature_names=feature_names,
            classification_threshold=classification_threshold,
            classification_labels=classification_labels,
            **kwargs,
        )

        self.huggingface_module = model.__class__
        self.pipeline_task = model.task if isinstance(model, pipelines.Pipeline) else None

    @classmethod
    def load_model(cls, local_path):
        huggingface_meta_file = Path(local_path) / "giskard-model-huggingface-meta.yaml"
        if huggingface_meta_file.exists():
            with open(huggingface_meta_file) as f:
                huggingface_meta = yaml.load(f, Loader=yaml.Loader)

        if huggingface_meta["pipeline_task"]:
            return pipeline(huggingface_meta["pipeline_task"], local_path)

        return huggingface_meta["huggingface_module"].from_pretrained(local_path)

    def save_huggingface_meta(self, local_path):
        with open(Path(local_path) / "giskard-model-huggingface-meta.yaml", "w") as f:
            yaml.dump(
                {
                    "huggingface_module": self.huggingface_module,
                    "pipeline_task": self.pipeline_task,
                },
                f,
                default_flow_style=False,
            )

    def save(self, local_path: Union[str, Path]) -> None:
        super().save(local_path)
        self.save_model(local_path)
        self.save_huggingface_meta(local_path)

    def save_model(self, local_path):
        self.model.save_pretrained(local_path)

    def model_predict(self, data):
        predictions = self._get_predictions(data)

        if self.is_classification and hasattr(predictions, "logits"):
            if isinstance(self.model, torch.nn.Module):
                with torch.no_grad():
                    logits = predictions.logits.detach().numpy()
            else:
                logits = predictions.logits

            if self.model_postprocessing_function:
                logger.warning(
                    "Your model output is logits. In Giskard, we expect the output to be probabilities."
                    "Since you provided a model_postprocessing_function, we assume that you included softmax() yourself.",
                    exc_info=True,
                )
            else:
                predictions = special.softmax(logits, axis=1) if len(logits.shape) == 2 else special.softmax(logits)

        return predictions

    def _get_predictions(self, data):
        if isinstance(self.model, torch.nn.Module):
            with torch.no_grad():
                return self.model(**data)

        if isinstance(self.model, pipelines.Pipeline):
            _predictions = [{p["label"]: p["score"] for p in pl} for pl in self.model(list(data), top_k=None)]
            return [[p[label] for label in self.meta.classification_labels] for p in _predictions]

        return self.model(**data)



import pandas as pd
from abc import abstractmethod
from dataclasses import dataclass
from functools import lru_cache

from ...testing.tests.calibration import test_underconfidence_rate

from ...testing.tests.calibration import test_overconfidence_rate

from ..common.examples import ExampleExtractor

from ..issues import Issue
from ...datasets.base import Dataset
from ...ml_worker.testing.registry.slicing_function import SlicingFunction
from ...models.base import BaseModel
from ...slicing.slice import QueryBasedSliceFunction
from ...slicing.text_slicer import MetadataSliceFunction


@dataclass
class CalibrationIssueInfo:
    slice_fn: SlicingFunction
    slice_size: int
    metric_value_slice: float
    metric_value_reference: float
    loss_values: pd.Series
    fail_idx: pd.DataFrame
    threshold: float
    p_threshold: float

    @property
    def metric_rel_delta(self):
        return (self.metric_value_slice - self.metric_value_reference) / self.metric_value_reference

    @property
    def metric_abs_delta(self):
        return self.metric_value_slice - self.metric_value_reference


class CalibrationIssue(Issue):
    group = "Calibration"

    info: CalibrationIssueInfo

    def __init__(
        self,
        model: BaseModel,
        dataset: Dataset,
        level: str,
        info: CalibrationIssueInfo,
    ):
        super().__init__(model, dataset, level, info)

    @property
    def domain(self):
        return str(self.info.slice_fn)

    @property
    @abstractmethod
    def metric(self):
        ...

    @property
    def deviation(self):
        return f"{self.info.metric_rel_delta * 100:.2f}% than global"

    @property
    def description(self):
        return f"{len(self.info.fail_idx)} out of {self.info.slice_size} samples"

    def _features(self):
        if isinstance(self.info.slice_fn, QueryBasedSliceFunction):
            return self.info.slice_fn.query.columns()
        if isinstance(self.info.slice_fn, MetadataSliceFunction):
            return [self.info.slice_fn.feature]
        return self.model.meta.feature_names or self.dataset.columns

    @property
    def importance(self):
        return self.info.metric_rel_delta

    @property
    def features(self):
        if isinstance(self.info.slice_fn, QueryBasedSliceFunction):
            return self.info.slice_fn.query.columns()
        if isinstance(self.info.slice_fn, MetadataSliceFunction):
            return [self.info.slice_fn.feature]
        return self.model.meta.feature_names or self.dataset.columns

    @lru_cache
    def examples(self, n=3):
        def filter_examples(issue, dataset):
            bad_pred_mask = dataset.df.index.isin(self.info.fail_idx)

            return dataset.slice(lambda df: df.loc[bad_pred_mask], row_level=False)

        def sort_examples(issue, examples):
            idx = self.info.loss_values.loc[examples.index].sort_values(ascending=False).index
            return examples.loc[idx]

        extractor = ExampleExtractor(self, filter_examples, sort_examples)
        return extractor.get_examples_dataframe(n, with_prediction=2)

    def generate_tests(self, with_names=False) -> list:
        return []


class OverconfidenceIssue(CalibrationIssue):
    group = "Overconfidence"

    @property
    def metric(self) -> str:
        return "Overconfidence rate"

    def generate_tests(self, with_names=False) -> list:
        abs_threshold = self.info.metric_value_reference * (1 + self.info.threshold)

        tests = [
            test_overconfidence_rate(
                model=self.model,
                dataset=self.dataset,
                slicing_function=self.info.slice_fn,
                threshold=abs_threshold,
                p_threshold=self.info.p_threshold,
            )
        ]

        if with_names:
            names = [f"Overconfidence on data slice “{self.info.slice_fn}”"]

            return list(zip(tests, names))

        return tests


class UnderconfidenceIssue(CalibrationIssue):
    group = "Underconfidence"

    _num_labels_display = 2

    @property
    def metric(self) -> str:
        return "Underconfidence rate"

    @property
    def deviation(self):
        return f"{self.info.metric_rel_delta * 100:.2f}% than global"

    def generate_tests(self, with_names=False) -> list:
        abs_threshold = self.info.metric_value_reference * (1 + self.info.threshold)

        tests = [
            test_underconfidence_rate(
                model=self.model,
                dataset=self.dataset,
                slicing_function=self.info.slice_fn,
                threshold=abs_threshold,
                p_threshold=self.info.p_threshold,
            )
        ]

        if with_names:
            names = [f"Underconfidence on data slice “{self.info.slice_fn}”"]

            return list(zip(tests, names))

        return tests



import numpy as np
import pytest

from giskard.models.base import BaseModel
from giskard.datasets.base import Dataset
from giskard.models.model_explanation import explain, explain_text


@pytest.mark.parametrize(
    "ds_name, model_name, include_feature_names",
    [
        ("german_credit_test_data", "german_credit_model", True),
        ("german_credit_data", "german_credit_model", True),
        ("diabetes_dataset", "linear_regression_diabetes", True),
        ("diabetes_dataset_with_target", "linear_regression_diabetes", True),
        ("german_credit_test_data", "german_credit_model", False),
        ("german_credit_data", "german_credit_model", False),
        ("diabetes_dataset", "linear_regression_diabetes", False),
        ("diabetes_dataset_with_target", "linear_regression_diabetes", False),
        ("enron_data", "enron_model", False),
    ],
)
def test_explain(ds_name: str, model_name: str, include_feature_names: bool, request):
    ds: Dataset = request.getfixturevalue(ds_name)
    model: BaseModel = request.getfixturevalue(model_name)

    # Try without feature names, it should also work
    if not include_feature_names:
        model.feature_names = None

    explanations = explain(model, ds, ds.df.iloc[0].to_dict())

    assert explanations and explanations.get("explanations")

    if model.is_classification:
        for label in model.meta.classification_labels:
            label_explanation = explanations.get("explanations").get(label)
            assert label_explanation
            for i, e in label_explanation.items():
                assert np.issubdtype(type(e), np.floating), f"'{i}' explanation value isn't float"
    elif model.is_regression:
        assert "default" in explanations.get("explanations")
        exp = explanations.get("explanations").get("default")
        for i, e in exp.items():
            assert np.issubdtype(type(e), np.floating), f"'{i}' explanation value isn't float"


def test_explain_shuffle_columns(german_credit_test_data, german_credit_model):
    german_credit_model.feature_names = None
    ds = german_credit_test_data
    # change column order
    res = explain(german_credit_model, ds, ds.df.iloc[0][ds.df.columns[::-1]].to_dict())
    assert res


def test_explain_text(enron_test_data, enron_model):
    df = enron_test_data.df.dropna()
    sample = df.head(1)
    result = explain_text(
        model=enron_model, input_df=sample, text_column="Content", text_document=sample["Content"].iloc[0], n_samples=10
    )
    assert result



import os
from pathlib import Path
from typing import Optional

from pydantic.env_settings import BaseSettings


def expand_env_var(env_var: Optional[str]) -> Optional[str]:
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


class Settings(BaseSettings):
    home: str = "~/giskard-home"
    port: int = 50051
    host: str = "localhost"
    max_workers: int = 10
    max_send_message_length_mb: int = 1024
    max_receive_message_length_mb: int = 1024
    loglevel = "INFO"
    cache_dir: str = "cache"
    disable_analytics = False

    class Config:
        env_prefix = "GSK_"

    @property
    def home_dir(self) -> Path:
        return Path(expand_env_var(self.home))


settings = Settings()



import asyncio
import logging
import os
import platform
import sys
import tempfile
import time
from io import StringIO
from pathlib import Path

import google
import grpc
import math
import numpy as np
import pandas as pd
import pkg_resources
import psutil
import tqdm

import giskard
from giskard.client.giskard_client import GiskardClient
from giskard.core.suite import Suite, ModelInput, DatasetInput, SuiteInput
from giskard.datasets.base import Dataset
from giskard.ml_worker.core.log_listener import LogListener
from giskard.ml_worker.exceptions.IllegalArgumentError import IllegalArgumentError
from giskard.ml_worker.exceptions.giskard_exception import GiskardException
from giskard.ml_worker.generated import ml_worker_pb2
from giskard.ml_worker.generated.ml_worker_pb2_grpc import MLWorkerServicer
from giskard.ml_worker.ml_worker import MLWorker
from giskard.ml_worker.testing.registry.giskard_test import GiskardTest
from giskard.ml_worker.testing.registry.registry import tests_registry
from giskard.ml_worker.testing.registry.slicing_function import SlicingFunction
from giskard.ml_worker.testing.registry.transformation_function import TransformationFunction
from giskard.ml_worker.testing.test_result import TestResult, TestMessageLevel
from giskard.ml_worker.utils.file_utils import get_file_name
from giskard.models.base import BaseModel
from giskard.models.model_explanation import (
    explain,
    explain_text,
)
from giskard.path_utils import model_path, dataset_path

logger = logging.getLogger(__name__)


def file_already_exists(meta: ml_worker_pb2.FileUploadMetadata):
    if meta.file_type == ml_worker_pb2.FileType.MODEL:
        path = model_path(meta.project_key, meta.name)
    elif meta.file_type == ml_worker_pb2.FileType.DATASET:
        path = dataset_path(meta.project_key, meta.name)
    else:
        raise ValueError(f"Illegal file type: {meta.file_type}")
    return path.exists(), path


def map_function_meta(callable_type):
    return {
        test.uuid: ml_worker_pb2.FunctionMeta(
            uuid=test.uuid,
            name=test.name,
            displayName=test.display_name,
            module=test.module,
            doc=test.doc,
            code=test.code,
            moduleDoc=test.module_doc,
            tags=test.tags,
            type=test.type,
            args=[
                ml_worker_pb2.TestFunctionArgument(
                    name=a.name, type=a.type, optional=a.optional, default=str(a.default), argOrder=a.argOrder
                )
                for a in test.args.values()
            ],
        )
        for test in tests_registry.get_all().values()
        if test.type == callable_type
    }


def map_dataset_process_function_meta(callable_type):
    return {
        test.uuid: ml_worker_pb2.DatasetProcessFunctionMeta(
            uuid=test.uuid,
            name=test.name,
            displayName=test.display_name,
            module=test.module,
            doc=test.doc,
            code=test.code,
            moduleDoc=test.module_doc,
            tags=test.tags,
            type=test.type,
            args=[
                ml_worker_pb2.TestFunctionArgument(
                    name=a.name, type=a.type, optional=a.optional, default=str(a.default), argOrder=a.argOrder
                )
                for a in test.args.values()
            ],
            cellLevel=test.cell_level,
            columnType=test.column_type,
            processType=test.process_type.name,
        )
        for test in tests_registry.get_all().values()
        if test.type == callable_type
    }


class MLWorkerServiceImpl(MLWorkerServicer):
    def __init__(
        self, ml_worker: MLWorker, client: GiskardClient, address=None, remote=None, loop=asyncio.get_event_loop()
    ) -> None:
        super().__init__()
        self.ml_worker = ml_worker
        self.address = address
        self.remote = remote
        self.client = client
        self.loop = loop

    def echo_orig(self, request, context):
        logger.debug(f"echo: {request.msg}")
        return ml_worker_pb2.EchoMsg(msg=request.msg)

    def upload(self, request_iterator, context: grpc.ServicerContext):
        meta = None
        path = None
        progress = None
        for upload_msg in request_iterator:
            if upload_msg.HasField("metadata"):
                meta = upload_msg.metadata
                file_exists, path = file_already_exists(meta)
                if not file_exists:
                    progress = tqdm.tqdm(
                        desc=f"Receiving {upload_msg.metadata.name}",
                        unit="B",
                        unit_scale=True,
                        unit_divisor=1024,
                    )
                    yield ml_worker_pb2.UploadStatus(code=ml_worker_pb2.StatusCode.CacheMiss)
                else:
                    logger.info(f"File already exists: {path}")
                    break
            elif upload_msg.HasField("chunk"):
                try:
                    path.parent.mkdir(exist_ok=True, parents=True)
                    with open(path, "ab") as f:
                        f.write(upload_msg.chunk.content)
                    progress.update(len(upload_msg.chunk.content))
                except Exception as e:
                    if progress is not None:
                        progress.close()
                    logger.exception(f"Failed to upload file {meta.name}", e)
                    yield ml_worker_pb2.UploadStatus(code=ml_worker_pb2.StatusCode.Failed)

        if progress is not None:
            progress.close()
        yield ml_worker_pb2.UploadStatus(code=ml_worker_pb2.StatusCode.Ok)

    def getInfo(self, request: ml_worker_pb2.MLWorkerInfoRequest, context):
        logger.info("Collecting ML Worker info")
        installed_packages = (
            {p.project_name: p.version for p in pkg_resources.working_set} if request.list_packages else None
        )
        current_process = psutil.Process(os.getpid())
        return ml_worker_pb2.MLWorkerInfo(
            platform=ml_worker_pb2.PlatformInfo(
                machine=platform.uname().machine,
                node=platform.uname().node,
                processor=platform.uname().processor,
                release=platform.uname().release,
                system=platform.uname().system,
                version=platform.uname().version,
            ),
            giskard_client_version=giskard.__version__,
            pid=os.getpid(),
            process_start_time=int(current_process.create_time()),
            interpreter=sys.executable,
            interpreter_version=platform.python_version(),
            installed_packages=installed_packages,
            internal_grpc_address=self.address,
            is_remote=self.remote,
        )

    def echo(self, request: ml_worker_pb2.EchoMsg, context: grpc.ServicerContext) -> ml_worker_pb2.EchoMsg:
        return request

    def runAdHocTest(
        self, request: ml_worker_pb2.RunAdHocTestRequest, context: grpc.ServicerContext
    ) -> ml_worker_pb2.TestResultMessage:
        test: GiskardTest = GiskardTest.download(request.testUuid, self.client, None)

        arguments = self.parse_function_arguments(request.arguments)

        logger.info(f"Executing {test.meta.display_name or f'{test.meta.module}.{test.meta.name}'}")
        test_result = test.get_builder()(**arguments).execute()

        return ml_worker_pb2.TestResultMessage(
            results=[
                ml_worker_pb2.NamedSingleTestResult(
                    testUuid=test.meta.uuid, result=map_result_to_single_test_result(test_result)
                )
            ]
        )

    def datasetProcessing(
        self, request: ml_worker_pb2.DatasetProcessingRequest, context: grpc.ServicerContext
    ) -> ml_worker_pb2.DatasetProcessingResultMessage:
        dataset = Dataset.download(self.client, request.dataset.project_key, request.dataset.id, request.dataset.sample)

        for function in request.functions:
            arguments = self.parse_function_arguments(function.arguments)
            if function.HasField("slicingFunction"):
                dataset.add_slicing_function(
                    SlicingFunction.download(function.slicingFunction.id, self.client, None)(**arguments)
                )
            else:
                dataset.add_transformation_function(
                    TransformationFunction.download(function.transformationFunction.id, self.client, None)(**arguments)
                )

        result = dataset.process()

        filtered_rows_idx = dataset.df.index.difference(result.df.index)
        modified_rows = result.df[dataset.df.iloc[result.df.index].ne(result.df)].dropna(how="all")

        return ml_worker_pb2.DatasetProcessingResultMessage(
            datasetId=request.dataset.id,
            totalRows=len(dataset.df.index),
            filteredRows=filtered_rows_idx,
            modifications=[
                ml_worker_pb2.DatasetRowModificationResult(
                    rowId=row[0],
                    modifications={
                        key: str(value)
                        for key, value in row[1].items()
                        if not type(value) == float or not math.isnan(value)
                    },
                )
                for row in modified_rows.iterrows()
            ],
        )

    def runTestSuite(
        self, request: ml_worker_pb2.RunTestSuiteRequest, context: grpc.ServicerContext
    ) -> ml_worker_pb2.TestSuiteResultMessage:
        log_listener = LogListener()
        try:
            tests = [
                {
                    "test": GiskardTest.download(t.testUuid, self.client, None),
                    "arguments": self.parse_function_arguments(t.arguments),
                    "id": t.id,
                }
                for t in request.tests
            ]

            global_arguments = self.parse_function_arguments(request.globalArguments)

            test_names = list(
                map(
                    lambda t: t["test"].meta.display_name or f"{t['test'].meta.module + '.' + t['test'].meta.name}",
                    tests,
                )
            )
            logger.info(f"Executing test suite: {test_names}")

            suite = Suite()
            for t in tests:
                suite.add_test(t["test"].get_builder()(**t["arguments"]), t["id"])

            is_pass, results = suite.run(**global_arguments)

            identifier_single_test_results = []
            for identifier, result in results:
                identifier_single_test_results.append(
                    ml_worker_pb2.IdentifierSingleTestResult(
                        id=identifier, result=map_result_to_single_test_result(result)
                    )
                )

            return ml_worker_pb2.TestSuiteResultMessage(
                is_error=False, is_pass=is_pass, results=identifier_single_test_results, logs=log_listener.close()
            )

        except Exception as exc:
            logger.exception("An error occurred during the test suite execution: %s", exc)
            return ml_worker_pb2.TestSuiteResultMessage(
                is_error=True, is_pass=False, results=[], logs=log_listener.close()
            )

    def parse_function_arguments(self, request_arguments):
        arguments = dict()

        for arg in request_arguments:
            if arg.none:
                continue
            if arg.HasField("dataset"):
                arguments[arg.name] = Dataset.download(
                    self.client, arg.dataset.project_key, arg.dataset.id, arg.dataset.sample
                )
            elif arg.HasField("model"):
                arguments[arg.name] = BaseModel.download(self.client, arg.model.project_key, arg.model.id)
            elif arg.HasField("slicingFunction"):
                arguments[arg.name] = SlicingFunction.download(arg.slicingFunction.id, self.client, None)(
                    **self.parse_function_arguments(arg.args)
                )
            elif arg.HasField("transformationFunction"):
                arguments[arg.name] = TransformationFunction.download(arg.transformationFunction.id, self.client, None)(
                    **self.parse_function_arguments(arg.args)
                )
            elif arg.HasField("float"):
                arguments[arg.name] = float(arg.float)
            elif arg.HasField("int"):
                arguments[arg.name] = int(arg.int)
            elif arg.HasField("str"):
                arguments[arg.name] = str(arg.str)
            elif arg.HasField("bool"):
                arguments[arg.name] = bool(arg.bool)
            elif arg.HasField("kwargs"):
                kwargs = dict()
                exec(arg.kwargs, {"kwargs": kwargs})
                arguments.update(kwargs)
            else:
                raise IllegalArgumentError("Unknown argument type")
        return arguments

    def explain(self, request: ml_worker_pb2.ExplainRequest, context) -> ml_worker_pb2.ExplainResponse:
        model = BaseModel.download(self.client, request.model.project_key, request.model.id)
        dataset = Dataset.download(self.client, request.dataset.project_key, request.dataset.id, request.dataset.sample)
        explanations = explain(model, dataset, request.columns)

        return ml_worker_pb2.ExplainResponse(
            explanations={
                str(k): ml_worker_pb2.ExplainResponse.Explanation(per_feature=v)
                for k, v in explanations["explanations"].items()
            }
        )

    def explainText(self, request: ml_worker_pb2.ExplainTextRequest, context) -> ml_worker_pb2.ExplainTextResponse:
        n_samples = 500 if request.n_samples <= 0 else request.n_samples
        model = BaseModel.download(self.client, request.model.project_key, request.model.id)
        text_column = request.feature_name

        if request.column_types[text_column] != "text":
            raise ValueError(f"Column {text_column} is not of type text")
        text_document = request.columns[text_column]
        input_df = pd.DataFrame({k: [v] for k, v in request.columns.items()})
        if model.meta.feature_names:
            input_df = input_df[model.meta.feature_names]
        (list_words, list_weights) = explain_text(model, input_df, text_column, text_document, n_samples)
        map_features_weight = dict(zip(model.meta.classification_labels, list_weights))
        return ml_worker_pb2.ExplainTextResponse(
            weights={
                str(k): ml_worker_pb2.ExplainTextResponse.WeightsPerFeature(
                    weights=[weight for weight in map_features_weight[k]]
                )
                for k in map_features_weight
            },
            words=list_words,
        )

    def runModelForDataFrame(self, request: ml_worker_pb2.RunModelForDataFrameRequest, context):
        model = BaseModel.download(self.client, request.model.project_key, request.model.id)
        df = pd.DataFrame.from_records([r.columns for r in request.dataframe.rows])
        ds = Dataset(
            model.prepare_dataframe(df, column_dtypes=request.column_dtypes),
            target=None,
            column_types=request.column_types,
        )
        predictions = model.predict(ds)
        if model.is_classification:
            return ml_worker_pb2.RunModelForDataFrameResponse(
                all_predictions=self.pandas_df_to_proto_df(predictions.all_predictions),
                prediction=predictions.prediction.astype(str),
            )
        else:
            return ml_worker_pb2.RunModelForDataFrameResponse(
                prediction=predictions.prediction.astype(str), raw_prediction=predictions.prediction
            )

    def runModel(self, request: ml_worker_pb2.RunModelRequest, context) -> ml_worker_pb2.RunModelResponse:
        try:
            model = BaseModel.download(self.client, request.model.project_key, request.model.id)
            dataset = Dataset.download(
                self.client, request.dataset.project_key, request.dataset.id, sample=request.dataset.sample
            )
        except ValueError as e:
            if "unsupported pickle protocol" in str(e):
                raise ValueError(
                    "Unable to unpickle object, "
                    "Make sure that Python version of client code is the same as the Python version in ML Worker."
                    "To change Python version, please refer to https://docs.giskard.ai/start/guides/configuration"
                    f"\nOriginal Error: {e}"
                ) from e
            raise e
        except ModuleNotFoundError as e:
            raise GiskardException(
                f"Failed to import '{e.name}'. "
                f"Make sure it's installed in the ML Worker environment."
                "To have more information on ML Worker, please see: https://docs.giskard.ai/start/guides/installation/ml-worker"
            ) from e
        prediction_results = model.predict(dataset)

        if model.is_classification:
            results = prediction_results.all_predictions
            labels = {k: v for k, v in enumerate(model.meta.classification_labels)}
            label_serie = dataset.df[dataset.target] if dataset.target else None
            if len(model.meta.classification_labels) > 2 or model.meta.classification_threshold is None:
                preds_serie = prediction_results.all_predictions.idxmax(axis="columns")
                sorted_predictions = np.sort(prediction_results.all_predictions.values)
                abs_diff = pd.Series(sorted_predictions[:, -1] - sorted_predictions[:, -2], name="absDiff")
            else:
                diff = prediction_results.all_predictions.iloc[:, 1] - model.meta.classification_threshold
                preds_serie = (diff >= 0).astype(int).map(labels).rename("predictions")
                abs_diff = pd.Series(diff.abs(), name="absDiff")
            calculated = pd.concat([preds_serie, label_serie, abs_diff], axis=1)
        else:
            results = pd.Series(prediction_results.prediction)
            preds_serie = results
            if dataset.target and dataset.target in dataset.df.columns:
                target_serie = dataset.df[dataset.target]
                diff = preds_serie - target_serie
                diff_percent = pd.Series(diff / target_serie, name="diffPercent")
                abs_diff = pd.Series(diff.abs(), name="absDiff")
                abs_diff_percent = pd.Series(abs_diff / target_serie, name="absDiffPercent")
                calculated = pd.concat([preds_serie, target_serie, abs_diff, abs_diff_percent, diff_percent], axis=1)
            else:
                calculated = pd.concat([preds_serie], axis=1)

        with tempfile.TemporaryDirectory(prefix="giskard-") as f:
            dir = Path(f)
            predictions_csv = get_file_name("predictions", "csv", request.dataset.sample)
            results.to_csv(index=False, path_or_buf=dir / predictions_csv)
            self.ml_worker.tunnel.client.log_artifact(
                dir / predictions_csv, f"{request.project_key}/models/inspections/{request.inspectionId}"
            )

            calculated_csv = get_file_name("calculated", "csv", request.dataset.sample)
            calculated.to_csv(index=False, path_or_buf=dir / calculated_csv)
            self.ml_worker.tunnel.client.log_artifact(
                dir / calculated_csv, f"{request.project_key}/models/inspections/{request.inspectionId}"
            )
        return google.protobuf.empty_pb2.Empty()

    def filterDataset(self, request_iterator, context: grpc.ServicerContext):
        filterfunc = {}
        meta = None

        times = []  # This is an array of chunk execution times for performance stats
        column_dtypes = []

        for filter_msg in request_iterator:
            if filter_msg.HasField("meta"):
                meta = filter_msg.meta
                try:
                    exec(meta.function, None, filterfunc)
                except Exception as e:
                    yield ml_worker_pb2.FilterDatasetResponse(
                        code=ml_worker_pb2.StatusCode.Failed, error_message=str(e)
                    )
                column_dtypes = meta.column_dtypes
                logger.info(f"Filtering dataset with {meta}")

                def filter_wrapper(row):
                    try:
                        return bool(filterfunc["filter_row"](row))
                    except Exception as e:  # noqa # NOSONAR
                        raise ValueError("Failed to execute user defined filtering function") from e

                yield ml_worker_pb2.FilterDatasetResponse(code=ml_worker_pb2.StatusCode.Ready)
            elif filter_msg.HasField("data"):
                logger.info("Got chunk " + str(filter_msg.idx))
                time_start = time.perf_counter()
                data_as_string = filter_msg.data.content.decode("utf-8")
                data_as_string = meta.headers + "\n" + data_as_string
                # CSV => Dataframe
                data = StringIO(data_as_string)  # Wrap using StringIO to avoid creating file
                df = pd.read_csv(data, keep_default_na=False, na_values=["_GSK_NA_"])
                df = df.astype(column_dtypes)
                # Iterate over rows, applying filter_row func
                try:
                    rows_to_keep = df[df.apply(filter_wrapper, axis=1)].index.array
                except Exception as e:
                    yield ml_worker_pb2.FilterDatasetResponse(
                        code=ml_worker_pb2.StatusCode.Failed, error_message=str(e)
                    )
                time_end = time.perf_counter()
                times.append(time_end - time_start)
                # Send NEXT code
                yield ml_worker_pb2.FilterDatasetResponse(
                    code=ml_worker_pb2.StatusCode.Next, idx=filter_msg.idx, rows=rows_to_keep
                )

        logger.info(f"Filter dataset finished. Avg chunk time: {sum(times) / len(times)}")
        yield ml_worker_pb2.FilterDatasetResponse(code=ml_worker_pb2.StatusCode.Ok)

    @staticmethod
    def map_suite_input(i: ml_worker_pb2.SuiteInput):
        if i.type == "Model" and i.model_meta is not None:
            return ModelInput(i.name, i.model_meta.model_type)
        elif i.type == "Dataset" and i.dataset_meta is not None:
            return DatasetInput(i.name, i.dataset_meta.target)
        else:
            return SuiteInput(i.name, i.type)

    def generateTestSuite(
        self,
        request: ml_worker_pb2.GenerateTestSuiteRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.GenerateTestSuiteResponse:
        inputs = [self.map_suite_input(i) for i in request.inputs]

        suite = Suite().generate_tests(inputs).to_dto(self.client, request.project_key)

        return ml_worker_pb2.GenerateTestSuiteResponse(
            tests=[
                ml_worker_pb2.GeneratedTest(
                    test_uuid=test.testUuid,
                    inputs=[
                        ml_worker_pb2.GeneratedTestInput(name=i.name, value=i.value, is_alias=i.is_alias)
                        for i in test.functionInputs.values()
                    ],
                )
                for test in suite.tests
            ]
        )

    def stopWorker(
        self, request: google.protobuf.empty_pb2.Empty, context: grpc.ServicerContext
    ) -> google.protobuf.empty_pb2.Empty:
        logger.info("Received request to stop the worker")
        self.loop.create_task(self.ml_worker.stop())
        return google.protobuf.empty_pb2.Empty()

    def getCatalog(
        self, request: google.protobuf.empty_pb2.Empty, context: grpc.ServicerContext
    ) -> ml_worker_pb2.CatalogResponse:
        return ml_worker_pb2.CatalogResponse(
            tests=map_function_meta("TEST"),
            slices=map_dataset_process_function_meta("SLICE"),
            transformations=map_dataset_process_function_meta("TRANSFORMATION"),
        )

    @staticmethod
    def pandas_df_to_proto_df(df):
        return ml_worker_pb2.DataFrame(
            rows=[
                ml_worker_pb2.DataRow(columns={str(k): v for k, v in r.astype(str).to_dict().items()})
                for _, r in df.iterrows()
            ]
        )

    @staticmethod
    def pandas_series_to_proto_series(self, series):
        return


def map_result_to_single_test_result(result) -> ml_worker_pb2.SingleTestResult:
    if isinstance(result, ml_worker_pb2.SingleTestResult):
        return result
    elif isinstance(result, TestResult):
        return ml_worker_pb2.SingleTestResult(
            passed=result.passed,
            is_error=result.is_error,
            messages=[
                ml_worker_pb2.TestMessage(
                    type=ml_worker_pb2.TestMessageType.ERROR
                    if message.type == TestMessageLevel.ERROR
                    else ml_worker_pb2.TestMessageType.INFO,
                    text=message.text,
                )
                for message in result.messages
            ]
            if result.messages is not None
            else [],
            props=result.props,
            metric=result.metric,
            missing_count=result.missing_count,
            missing_percent=result.missing_percent,
            unexpected_count=result.unexpected_count,
            unexpected_percent=result.unexpected_percent,
            unexpected_percent_total=result.unexpected_percent_total,
            unexpected_percent_nonmissing=result.unexpected_percent_nonmissing,
            partial_unexpected_index_list=[
                ml_worker_pb2.Partial_unexpected_counts(value=puc.value, count=puc.count)
                for puc in result.partial_unexpected_index_list
            ],
            unexpected_index_list=result.unexpected_index_list,
            output_df=result.output_df,
            number_of_perturbed_rows=result.number_of_perturbed_rows,
            actual_slices_size=result.actual_slices_size,
            reference_slices_size=result.reference_slices_size,
        )
    elif isinstance(result, bool):
        return ml_worker_pb2.SingleTestResult(passed=result)
    else:
        raise ValueError("Result of test can only be 'TestResult' or 'bool'")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ml-worker.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fml-worker.proto\x12\x06worker\x1a\x19google/protobuf/any.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x1bgoogle/protobuf/empty.proto\",\n\x13MLWorkerInfoRequest\x12\x15\n\rlist_packages\x18\x01 \x01(\x08\"\xe6\x02\n\x0cMLWorkerInfo\x12&\n\x08platform\x18\x01 \x01(\x0b\x32\x14.worker.PlatformInfo\x12\x13\n\x0binterpreter\x18\x02 \x01(\t\x12\x1b\n\x13interpreter_version\x18\x03 \x01(\t\x12G\n\x12installed_packages\x18\x04 \x03(\x0b\x32+.worker.MLWorkerInfo.InstalledPackagesEntry\x12\x1d\n\x15internal_grpc_address\x18\x05 \x01(\t\x12\x11\n\tis_remote\x18\x06 \x01(\x08\x12\x0b\n\x03pid\x18\x07 \x01(\r\x12\x1a\n\x12process_start_time\x18\x08 \x01(\x04\x12\x1e\n\x16giskard_client_version\x18\t \x01(\t\x1a\x38\n\x16InstalledPackagesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"r\n\x0cPlatformInfo\x12\x0f\n\x07machine\x18\x01 \x01(\t\x12\x0c\n\x04node\x18\x02 \x01(\t\x12\x11\n\tprocessor\x18\x03 \x01(\t\x12\x0f\n\x07release\x18\x04 \x01(\t\x12\x0e\n\x06system\x18\x05 \x01(\t\x12\x0f\n\x07version\x18\x06 \x01(\t\"n\n\x11\x46ileUploadRequest\x12.\n\x08metadata\x18\x01 \x01(\x0b\x32\x1a.worker.FileUploadMetadataH\x00\x12\x1e\n\x05\x63hunk\x18\x02 \x01(\x0b\x32\r.worker.ChunkH\x00\x42\t\n\x07request\"h\n\x12\x46ileUploadMetadata\x12\n\n\x02id\x18\x01 \x01(\x04\x12#\n\tfile_type\x18\x02 \x01(\x0e\x32\x10.worker.FileType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x13\n\x0bproject_key\x18\x04 \x01(\t\"\x16\n\x07\x45\x63hoMsg\x12\x0b\n\x03msg\x18\x01 \x01(\t\"\xc0\x01\n\x0e\x45xplainRequest\x12\"\n\x05model\x18\x01 \x01(\x0b\x32\x13.worker.ArtifactRef\x12$\n\x07\x64\x61taset\x18\x02 \x01(\x0b\x32\x13.worker.ArtifactRef\x12\x34\n\x07\x63olumns\x18\x03 \x03(\x0b\x32#.worker.ExplainRequest.ColumnsEntry\x1a.\n\x0c\x43olumnsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc2\x02\n\x12\x45xplainTextRequest\x12\"\n\x05model\x18\x01 \x01(\x0b\x32\x13.worker.ArtifactRef\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x02 \x01(\t\x12\x38\n\x07\x63olumns\x18\x03 \x03(\x0b\x32\'.worker.ExplainTextRequest.ColumnsEntry\x12\x41\n\x0c\x63olumn_types\x18\x04 \x03(\x0b\x32+.worker.ExplainTextRequest.ColumnTypesEntry\x12\x11\n\tn_samples\x18\x05 \x01(\r\x1a.\n\x0c\x43olumnsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x32\n\x10\x43olumnTypesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xb9\x02\n\x0f\x45xplainResponse\x12?\n\x0c\x65xplanations\x18\x01 \x03(\x0b\x32).worker.ExplainResponse.ExplanationsEntry\x1a\x8a\x01\n\x0b\x45xplanation\x12H\n\x0bper_feature\x18\x01 \x03(\x0b\x32\x33.worker.ExplainResponse.Explanation.PerFeatureEntry\x1a\x31\n\x0fPerFeatureEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1aX\n\x11\x45xplanationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.worker.ExplainResponse.Explanation:\x02\x38\x01\"\xe4\x01\n\x13\x45xplainTextResponse\x12\r\n\x05words\x18\x01 \x03(\t\x12\x39\n\x07weights\x18\x02 \x03(\x0b\x32(.worker.ExplainTextResponse.WeightsEntry\x1a$\n\x11WeightsPerFeature\x12\x0f\n\x07weights\x18\x01 \x03(\x02\x1a]\n\x0cWeightsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12<\n\x05value\x18\x02 \x01(\x0b\x32-.worker.ExplainTextResponse.WeightsPerFeature:\x02\x38\x01\"\x8d\x01\n\x1cRunModelForDataFrameResponse\x12*\n\x0f\x61ll_predictions\x18\x01 \x01(\x0b\x32\x11.worker.DataFrame\x12\x12\n\nprediction\x18\x02 \x03(\t\x12\x15\n\rprobabilities\x18\x03 \x03(\x02\x12\x16\n\x0eraw_prediction\x18\x04 \x03(\x02\"h\n\x07\x44\x61taRow\x12-\n\x07\x63olumns\x18\x01 \x03(\x0b\x32\x1c.worker.DataRow.ColumnsEntry\x1a.\n\x0c\x43olumnsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"*\n\tDataFrame\x12\x1d\n\x04rows\x18\x01 \x03(\x0b\x32\x0f.worker.DataRow\"\xfa\x02\n\x1bRunModelForDataFrameRequest\x12\"\n\x05model\x18\x01 \x01(\x0b\x32\x13.worker.ArtifactRef\x12$\n\tdataframe\x18\x02 \x01(\x0b\x32\x11.worker.DataFrame\x12\x0e\n\x06target\x18\x03 \x01(\t\x12J\n\x0c\x63olumn_types\x18\x04 \x03(\x0b\x32\x34.worker.RunModelForDataFrameRequest.ColumnTypesEntry\x12L\n\rcolumn_dtypes\x18\x05 \x03(\x0b\x32\x35.worker.RunModelForDataFrameRequest.ColumnDtypesEntry\x1a\x32\n\x10\x43olumnTypesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x33\n\x11\x43olumnDtypesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x86\x01\n\x0fRunModelRequest\x12\"\n\x05model\x18\x01 \x01(\x0b\x32\x13.worker.ArtifactRef\x12$\n\x07\x64\x61taset\x18\x02 \x01(\x0b\x32\x13.worker.ArtifactRef\x12\x14\n\x0cinspectionId\x18\x03 \x01(\x04\x12\x13\n\x0bproject_key\x18\x04 \x01(\t\"?\n\x10RunModelResponse\x12\x13\n\x0bresults_csv\x18\x01 \x01(\t\x12\x16\n\x0e\x63\x61lculated_csv\x18\x02 \x01(\t\"P\n\x13RunAdHocTestRequest\x12\x10\n\x08testUuid\x18\x01 \x01(\t\x12\'\n\targuments\x18\x02 \x03(\x0b\x32\x14.worker.FuncArgument\"v\n\x18\x44\x61tasetProcessingRequest\x12$\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x13.worker.ArtifactRef\x12\x34\n\tfunctions\x18\x02 \x03(\x0b\x32!.worker.DatasetProcessingFunction\"\xb7\x01\n\x19\x44\x61tasetProcessingFunction\x12.\n\x0fslicingFunction\x18\x01 \x01(\x0b\x32\x13.worker.ArtifactRefH\x00\x12\x35\n\x16transformationFunction\x18\x02 \x01(\x0b\x32\x13.worker.ArtifactRefH\x00\x12\'\n\targuments\x18\x03 \x03(\x0b\x32\x14.worker.FuncArgumentB\n\n\x08\x66unction\"Z\n\x11SuiteTestArgument\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x10\n\x08testUuid\x18\x02 \x01(\t\x12\'\n\targuments\x18\x03 \x03(\x0b\x32\x14.worker.FuncArgument\"n\n\x13RunTestSuiteRequest\x12(\n\x05tests\x18\x01 \x03(\x0b\x32\x19.worker.SuiteTestArgument\x12-\n\x0fglobalArguments\x18\x02 \x03(\x0b\x32\x14.worker.FuncArgument\"\xe0\x02\n\x0c\x46uncArgument\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x05model\x18\x02 \x01(\x0b\x32\x13.worker.ArtifactRefH\x00\x12&\n\x07\x64\x61taset\x18\x03 \x01(\x0b\x32\x13.worker.ArtifactRefH\x00\x12\x0f\n\x05\x66loat\x18\x04 \x01(\x02H\x00\x12\r\n\x03int\x18\x05 \x01(\x05H\x00\x12\r\n\x03str\x18\x06 \x01(\tH\x00\x12\x0e\n\x04\x62ool\x18\x07 \x01(\x08H\x00\x12.\n\x0fslicingFunction\x18\x08 \x01(\x0b\x32\x13.worker.ArtifactRefH\x00\x12\x35\n\x16transformationFunction\x18\t \x01(\x0b\x32\x13.worker.ArtifactRefH\x00\x12\x10\n\x06kwargs\x18\n \x01(\tH\x00\x12\"\n\x04\x61rgs\x18\x0b \x03(\x0b\x32\x14.worker.FuncArgument\x12\x0c\n\x04none\x18\x0c \x01(\x08\x42\n\n\x08\x61rgument\"\x1f\n\x0fRunTestResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\"9\n\x19Partial_unexpected_counts\x12\r\n\x05value\x18\x01 \x03(\r\x12\r\n\x05\x63ount\x18\x02 \x01(\r\"S\n\x15NamedSingleTestResult\x12\x10\n\x08testUuid\x18\x01 \x01(\t\x12(\n\x06result\x18\x02 \x01(\x0b\x32\x18.worker.SingleTestResult\"R\n\x1aIdentifierSingleTestResult\x12\n\n\x02id\x18\x01 \x01(\x03\x12(\n\x06result\x18\x02 \x01(\x0b\x32\x18.worker.SingleTestResult\"B\n\x0bTestMessage\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x17.worker.TestMessageType\x12\x0c\n\x04text\x18\x02 \x01(\t\"\xf8\x04\n\x10SingleTestResult\x12\x0e\n\x06passed\x18\r \x01(\x08\x12\x10\n\x08is_error\x18\x17 \x01(\x08\x12%\n\x08messages\x18\x10 \x03(\x0b\x32\x13.worker.TestMessage\x12\x32\n\x05props\x18\x0e \x03(\x0b\x32#.worker.SingleTestResult.PropsEntry\x12\x0e\n\x06metric\x18\x0f \x01(\x02\x12\x15\n\rmissing_count\x18\x02 \x01(\x05\x12\x17\n\x0fmissing_percent\x18\x03 \x01(\x01\x12\x18\n\x10unexpected_count\x18\x04 \x01(\x05\x12\x1a\n\x12unexpected_percent\x18\x05 \x01(\x01\x12 \n\x18unexpected_percent_total\x18\x06 \x01(\x01\x12%\n\x1dunexpected_percent_nonmissing\x18\x07 \x01(\x01\x12%\n\x1dpartial_unexpected_index_list\x18\t \x03(\r\x12\x44\n\x19partial_unexpected_counts\x18\n \x03(\x0b\x32!.worker.Partial_unexpected_counts\x12\x1d\n\x15unexpected_index_list\x18\x0c \x03(\r\x12\x11\n\toutput_df\x18\x12 \x01(\x0c\x12 \n\x18number_of_perturbed_rows\x18\x14 \x01(\r\x12\x1a\n\x12\x61\x63tual_slices_size\x18\x15 \x03(\r\x12\x1d\n\x15reference_slices_size\x18\x16 \x03(\r\x1a,\n\nPropsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"C\n\x11TestResultMessage\x12.\n\x07results\x18\x01 \x03(\x0b\x32\x1d.worker.NamedSingleTestResult\"\x99\x01\n\x1e\x44\x61tasetProcessingResultMessage\x12\x11\n\tdatasetId\x18\x01 \x01(\t\x12\x11\n\ttotalRows\x18\x02 \x01(\r\x12\x14\n\x0c\x66ilteredRows\x18\x03 \x03(\r\x12;\n\rmodifications\x18\x04 \x03(\x0b\x32$.worker.DatasetRowModificationResult\"\xb3\x01\n\x1c\x44\x61tasetRowModificationResult\x12\r\n\x05rowId\x18\x01 \x01(\r\x12N\n\rmodifications\x18\x02 \x03(\x0b\x32\x37.worker.DatasetRowModificationResult.ModificationsEntry\x1a\x34\n\x12ModificationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"~\n\x16TestSuiteResultMessage\x12\x10\n\x08is_error\x18\x01 \x01(\x08\x12\x0f\n\x07is_pass\x18\x02 \x01(\x08\x12\x33\n\x07results\x18\x03 \x03(\x0b\x32\".worker.IdentifierSingleTestResult\x12\x0c\n\x04logs\x18\x04 \x01(\t\"N\n\x0b\x41rtifactRef\x12\x13\n\x0bproject_key\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x13\n\x06sample\x18\x03 \x01(\x08H\x00\x88\x01\x01\x42\t\n\x07_sample\"\x18\n\x05\x43hunk\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\"0\n\x0cUploadStatus\x12 \n\x04\x63ode\x18\x01 \x01(\x0e\x32\x12.worker.StatusCode\"1\n\x11MLWorkerErrorInfo\x12\r\n\x05\x65rror\x18\x01 \x01(\t\x12\r\n\x05stack\x18\x02 \x01(\t\"\xa4\x01\n\nSuiteInput\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12*\n\nmodel_meta\x18\x03 \x01(\x0b\x32\x11.worker.ModelMetaH\x00\x88\x01\x01\x12.\n\x0c\x64\x61taset_meta\x18\x04 \x01(\x0b\x32\x13.worker.DatasetMetaH\x01\x88\x01\x01\x42\r\n\x0b_model_metaB\x0f\n\r_dataset_meta\"3\n\tModelMeta\x12\x17\n\nmodel_type\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\r\n\x0b_model_type\"-\n\x0b\x44\x61tasetMeta\x12\x13\n\x06target\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_target\"S\n\x18GenerateTestSuiteRequest\x12\x13\n\x0bproject_key\x18\x01 \x01(\t\x12\"\n\x06inputs\x18\x02 \x03(\x0b\x32\x12.worker.SuiteInput\"A\n\x19GenerateTestSuiteResponse\x12$\n\x05tests\x18\x01 \x03(\x0b\x32\x15.worker.GeneratedTest\"N\n\rGeneratedTest\x12\x11\n\ttest_uuid\x18\x01 \x01(\t\x12*\n\x06inputs\x18\x02 \x03(\x0b\x32\x1a.worker.GeneratedTestInput\"C\n\x12GeneratedTestInput\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x10\n\x08is_alias\x18\x03 \x01(\x08\"\xb3\x03\n\x0f\x43\x61talogResponse\x12\x31\n\x05tests\x18\x01 \x03(\x0b\x32\".worker.CatalogResponse.TestsEntry\x12\x33\n\x06slices\x18\x02 \x03(\x0b\x32#.worker.CatalogResponse.SlicesEntry\x12\x45\n\x0ftransformations\x18\x03 \x03(\x0b\x32,.worker.CatalogResponse.TransformationsEntry\x1a\x42\n\nTestsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.worker.FunctionMeta:\x02\x38\x01\x1aQ\n\x0bSlicesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x31\n\x05value\x18\x02 \x01(\x0b\x32\".worker.DatasetProcessFunctionMeta:\x02\x38\x01\x1aZ\n\x14TransformationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x31\n\x05value\x18\x02 \x01(\x0b\x32\".worker.DatasetProcessFunctionMeta:\x02\x38\x01\"\xd6\x01\n\x0c\x46unctionMeta\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64isplayName\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\r\x12\x0e\n\x06module\x18\x05 \x01(\t\x12\x0b\n\x03\x64oc\x18\x06 \x01(\t\x12\x11\n\tmoduleDoc\x18\x07 \x01(\t\x12*\n\x04\x61rgs\x18\x08 \x03(\x0b\x32\x1c.worker.TestFunctionArgument\x12\x0c\n\x04tags\x18\t \x03(\t\x12\x0c\n\x04\x63ode\x18\n \x01(\t\x12\x0c\n\x04type\x18\x0b \x01(\t\"\xb4\x02\n\x1a\x44\x61tasetProcessFunctionMeta\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64isplayName\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\r\x12\x0e\n\x06module\x18\x05 \x01(\t\x12\x0b\n\x03\x64oc\x18\x06 \x01(\t\x12\x11\n\tmoduleDoc\x18\x07 \x01(\t\x12*\n\x04\x61rgs\x18\x08 \x03(\x0b\x32\x1c.worker.TestFunctionArgument\x12\x0c\n\x04tags\x18\t \x03(\t\x12\x0c\n\x04\x63ode\x18\n \x01(\t\x12\x0c\n\x04type\x18\x0b \x01(\t\x12\x11\n\tcellLevel\x18\x0c \x01(\x08\x12\x17\n\ncolumnType\x18\r \x01(\tH\x00\x88\x01\x01\x12\x13\n\x0bprocessType\x18\x0e \x01(\tB\r\n\x0b_columnType\"g\n\x14TestFunctionArgument\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x10\n\x08optional\x18\x03 \x01(\x08\x12\x0f\n\x07\x64\x65\x66\x61ult\x18\x04 \x01(\t\x12\x10\n\x08\x61rgOrder\x18\x05 \x01(\r\"\x8c\x01\n\x10\x43omparisonClause\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12/\n\x0f\x63omparison_type\x18\x02 \x01(\x0e\x32\x16.worker.ComparisonType\x12\x14\n\x0c\x63olumn_dtype\x18\x03 \x01(\t\x12\x12\n\x05value\x18\x04 \x01(\tH\x00\x88\x01\x01\x42\x08\n\x06_value\"V\n\x1eGenerateQueryBasedSliceRequest\x12\x34\n\x12\x63omparison_clauses\x18\x01 \x03(\x0b\x32\x18.worker.ComparisonClause\"-\n\x1fGenerateQueryBasedSliceResponse\x12\n\n\x02id\x18\x01 \x01(\t*\"\n\x08\x46ileType\x12\t\n\x05MODEL\x10\x00\x12\x0b\n\x07\x44\x41TASET\x10\x01*&\n\x0fTestMessageType\x12\t\n\x05\x45RROR\x10\x00\x12\x08\n\x04INFO\x10\x01*Q\n\nStatusCode\x12\x0b\n\x07Unknown\x10\x00\x12\x06\n\x02Ok\x10\x01\x12\n\n\x06\x46\x61iled\x10\x02\x12\r\n\tCacheMiss\x10\x03\x12\t\n\x05Ready\x10\x04\x12\x08\n\x04Next\x10\x05*\x89\x01\n\x0e\x43omparisonType\x12\x06\n\x02IS\x10\x00\x12\n\n\x06IS_NOT\x10\x01\x12\x0c\n\x08\x43ONTAINS\x10\x02\x12\x15\n\x11\x44OES_NOT_CONTAINS\x10\x03\x12\x0f\n\x0bSTARTS_WITH\x10\x04\x12\r\n\tENDS_WITH\x10\x05\x12\x0c\n\x08IS_EMPTY\x10\x06\x12\x10\n\x0cIS_NOT_EMPTY\x10\x07\x32\xf1\x07\n\x08MLWorker\x12>\n\x07getInfo\x12\x1b.worker.MLWorkerInfoRequest\x1a\x14.worker.MLWorkerInfo\"\x00\x12H\n\x0crunAdHocTest\x12\x1b.worker.RunAdHocTestRequest\x1a\x19.worker.TestResultMessage\"\x00\x12_\n\x11\x64\x61tasetProcessing\x12 .worker.DatasetProcessingRequest\x1a&.worker.DatasetProcessingResultMessage\"\x00\x12M\n\x0crunTestSuite\x12\x1b.worker.RunTestSuiteRequest\x1a\x1e.worker.TestSuiteResultMessage\"\x00\x12=\n\x08runModel\x12\x17.worker.RunModelRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x63\n\x14runModelForDataFrame\x12#.worker.RunModelForDataFrameRequest\x1a$.worker.RunModelForDataFrameResponse\"\x00\x12<\n\x07\x65xplain\x12\x16.worker.ExplainRequest\x1a\x17.worker.ExplainResponse\"\x00\x12H\n\x0b\x65xplainText\x12\x1a.worker.ExplainTextRequest\x1a\x1b.worker.ExplainTextResponse\"\x00\x12*\n\x04\x65\x63ho\x12\x0f.worker.EchoMsg\x1a\x0f.worker.EchoMsg\"\x00\x12Z\n\x11generateTestSuite\x12 .worker.GenerateTestSuiteRequest\x1a!.worker.GenerateTestSuiteResponse\"\x00\x12>\n\nstopWorker\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x00\x12?\n\ngetCatalog\x12\x16.google.protobuf.Empty\x1a\x17.worker.CatalogResponse\"\x00\x12v\n!generateQueryBasedSlicingFunction\x12&.worker.GenerateQueryBasedSliceRequest\x1a\'.worker.GenerateQueryBasedSliceResponse\"\x00\x42(\n\x11\x61i.giskard.workerB\x0bWorkerProtoP\x01\xa2\x02\x03WRKb\x06proto3')

_FILETYPE = DESCRIPTOR.enum_types_by_name['FileType']
FileType = enum_type_wrapper.EnumTypeWrapper(_FILETYPE)
_TESTMESSAGETYPE = DESCRIPTOR.enum_types_by_name['TestMessageType']
TestMessageType = enum_type_wrapper.EnumTypeWrapper(_TESTMESSAGETYPE)
_STATUSCODE = DESCRIPTOR.enum_types_by_name['StatusCode']
StatusCode = enum_type_wrapper.EnumTypeWrapper(_STATUSCODE)
_COMPARISONTYPE = DESCRIPTOR.enum_types_by_name['ComparisonType']
ComparisonType = enum_type_wrapper.EnumTypeWrapper(_COMPARISONTYPE)
MODEL = 0
DATASET = 1
ERROR = 0
INFO = 1
Unknown = 0
Ok = 1
Failed = 2
CacheMiss = 3
Ready = 4
Next = 5
IS = 0
IS_NOT = 1
CONTAINS = 2
DOES_NOT_CONTAINS = 3
STARTS_WITH = 4
ENDS_WITH = 5
IS_EMPTY = 6
IS_NOT_EMPTY = 7


_MLWORKERINFOREQUEST = DESCRIPTOR.message_types_by_name['MLWorkerInfoRequest']
_MLWORKERINFO = DESCRIPTOR.message_types_by_name['MLWorkerInfo']
_MLWORKERINFO_INSTALLEDPACKAGESENTRY = _MLWORKERINFO.nested_types_by_name['InstalledPackagesEntry']
_PLATFORMINFO = DESCRIPTOR.message_types_by_name['PlatformInfo']
_FILEUPLOADREQUEST = DESCRIPTOR.message_types_by_name['FileUploadRequest']
_FILEUPLOADMETADATA = DESCRIPTOR.message_types_by_name['FileUploadMetadata']
_ECHOMSG = DESCRIPTOR.message_types_by_name['EchoMsg']
_EXPLAINREQUEST = DESCRIPTOR.message_types_by_name['ExplainRequest']
_EXPLAINREQUEST_COLUMNSENTRY = _EXPLAINREQUEST.nested_types_by_name['ColumnsEntry']
_EXPLAINTEXTREQUEST = DESCRIPTOR.message_types_by_name['ExplainTextRequest']
_EXPLAINTEXTREQUEST_COLUMNSENTRY = _EXPLAINTEXTREQUEST.nested_types_by_name['ColumnsEntry']
_EXPLAINTEXTREQUEST_COLUMNTYPESENTRY = _EXPLAINTEXTREQUEST.nested_types_by_name['ColumnTypesEntry']
_EXPLAINRESPONSE = DESCRIPTOR.message_types_by_name['ExplainResponse']
_EXPLAINRESPONSE_EXPLANATION = _EXPLAINRESPONSE.nested_types_by_name['Explanation']
_EXPLAINRESPONSE_EXPLANATION_PERFEATUREENTRY = _EXPLAINRESPONSE_EXPLANATION.nested_types_by_name['PerFeatureEntry']
_EXPLAINRESPONSE_EXPLANATIONSENTRY = _EXPLAINRESPONSE.nested_types_by_name['ExplanationsEntry']
_EXPLAINTEXTRESPONSE = DESCRIPTOR.message_types_by_name['ExplainTextResponse']
_EXPLAINTEXTRESPONSE_WEIGHTSPERFEATURE = _EXPLAINTEXTRESPONSE.nested_types_by_name['WeightsPerFeature']
_EXPLAINTEXTRESPONSE_WEIGHTSENTRY = _EXPLAINTEXTRESPONSE.nested_types_by_name['WeightsEntry']
_RUNMODELFORDATAFRAMERESPONSE = DESCRIPTOR.message_types_by_name['RunModelForDataFrameResponse']
_DATAROW = DESCRIPTOR.message_types_by_name['DataRow']
_DATAROW_COLUMNSENTRY = _DATAROW.nested_types_by_name['ColumnsEntry']
_DATAFRAME = DESCRIPTOR.message_types_by_name['DataFrame']
_RUNMODELFORDATAFRAMEREQUEST = DESCRIPTOR.message_types_by_name['RunModelForDataFrameRequest']
_RUNMODELFORDATAFRAMEREQUEST_COLUMNTYPESENTRY = _RUNMODELFORDATAFRAMEREQUEST.nested_types_by_name['ColumnTypesEntry']
_RUNMODELFORDATAFRAMEREQUEST_COLUMNDTYPESENTRY = _RUNMODELFORDATAFRAMEREQUEST.nested_types_by_name['ColumnDtypesEntry']
_RUNMODELREQUEST = DESCRIPTOR.message_types_by_name['RunModelRequest']
_RUNMODELRESPONSE = DESCRIPTOR.message_types_by_name['RunModelResponse']
_RUNADHOCTESTREQUEST = DESCRIPTOR.message_types_by_name['RunAdHocTestRequest']
_DATASETPROCESSINGREQUEST = DESCRIPTOR.message_types_by_name['DatasetProcessingRequest']
_DATASETPROCESSINGFUNCTION = DESCRIPTOR.message_types_by_name['DatasetProcessingFunction']
_SUITETESTARGUMENT = DESCRIPTOR.message_types_by_name['SuiteTestArgument']
_RUNTESTSUITEREQUEST = DESCRIPTOR.message_types_by_name['RunTestSuiteRequest']
_FUNCARGUMENT = DESCRIPTOR.message_types_by_name['FuncArgument']
_RUNTESTRESPONSE = DESCRIPTOR.message_types_by_name['RunTestResponse']
_PARTIAL_UNEXPECTED_COUNTS = DESCRIPTOR.message_types_by_name['Partial_unexpected_counts']
_NAMEDSINGLETESTRESULT = DESCRIPTOR.message_types_by_name['NamedSingleTestResult']
_IDENTIFIERSINGLETESTRESULT = DESCRIPTOR.message_types_by_name['IdentifierSingleTestResult']
_TESTMESSAGE = DESCRIPTOR.message_types_by_name['TestMessage']
_SINGLETESTRESULT = DESCRIPTOR.message_types_by_name['SingleTestResult']
_SINGLETESTRESULT_PROPSENTRY = _SINGLETESTRESULT.nested_types_by_name['PropsEntry']
_TESTRESULTMESSAGE = DESCRIPTOR.message_types_by_name['TestResultMessage']
_DATASETPROCESSINGRESULTMESSAGE = DESCRIPTOR.message_types_by_name['DatasetProcessingResultMessage']
_DATASETROWMODIFICATIONRESULT = DESCRIPTOR.message_types_by_name['DatasetRowModificationResult']
_DATASETROWMODIFICATIONRESULT_MODIFICATIONSENTRY = _DATASETROWMODIFICATIONRESULT.nested_types_by_name['ModificationsEntry']
_TESTSUITERESULTMESSAGE = DESCRIPTOR.message_types_by_name['TestSuiteResultMessage']
_ARTIFACTREF = DESCRIPTOR.message_types_by_name['ArtifactRef']
_CHUNK = DESCRIPTOR.message_types_by_name['Chunk']
_UPLOADSTATUS = DESCRIPTOR.message_types_by_name['UploadStatus']
_MLWORKERERRORINFO = DESCRIPTOR.message_types_by_name['MLWorkerErrorInfo']
_SUITEINPUT = DESCRIPTOR.message_types_by_name['SuiteInput']
_MODELMETA = DESCRIPTOR.message_types_by_name['ModelMeta']
_DATASETMETA = DESCRIPTOR.message_types_by_name['DatasetMeta']
_GENERATETESTSUITEREQUEST = DESCRIPTOR.message_types_by_name['GenerateTestSuiteRequest']
_GENERATETESTSUITERESPONSE = DESCRIPTOR.message_types_by_name['GenerateTestSuiteResponse']
_GENERATEDTEST = DESCRIPTOR.message_types_by_name['GeneratedTest']
_GENERATEDTESTINPUT = DESCRIPTOR.message_types_by_name['GeneratedTestInput']
_CATALOGRESPONSE = DESCRIPTOR.message_types_by_name['CatalogResponse']
_CATALOGRESPONSE_TESTSENTRY = _CATALOGRESPONSE.nested_types_by_name['TestsEntry']
_CATALOGRESPONSE_SLICESENTRY = _CATALOGRESPONSE.nested_types_by_name['SlicesEntry']
_CATALOGRESPONSE_TRANSFORMATIONSENTRY = _CATALOGRESPONSE.nested_types_by_name['TransformationsEntry']
_FUNCTIONMETA = DESCRIPTOR.message_types_by_name['FunctionMeta']
_DATASETPROCESSFUNCTIONMETA = DESCRIPTOR.message_types_by_name['DatasetProcessFunctionMeta']
_TESTFUNCTIONARGUMENT = DESCRIPTOR.message_types_by_name['TestFunctionArgument']
_COMPARISONCLAUSE = DESCRIPTOR.message_types_by_name['ComparisonClause']
_GENERATEQUERYBASEDSLICEREQUEST = DESCRIPTOR.message_types_by_name['GenerateQueryBasedSliceRequest']
_GENERATEQUERYBASEDSLICERESPONSE = DESCRIPTOR.message_types_by_name['GenerateQueryBasedSliceResponse']
MLWorkerInfoRequest = _reflection.GeneratedProtocolMessageType('MLWorkerInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _MLWORKERINFOREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.MLWorkerInfoRequest)
  })
_sym_db.RegisterMessage(MLWorkerInfoRequest)

MLWorkerInfo = _reflection.GeneratedProtocolMessageType('MLWorkerInfo', (_message.Message,), {

  'InstalledPackagesEntry' : _reflection.GeneratedProtocolMessageType('InstalledPackagesEntry', (_message.Message,), {
    'DESCRIPTOR' : _MLWORKERINFO_INSTALLEDPACKAGESENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.MLWorkerInfo.InstalledPackagesEntry)
    })
  ,
  'DESCRIPTOR' : _MLWORKERINFO,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.MLWorkerInfo)
  })
_sym_db.RegisterMessage(MLWorkerInfo)
_sym_db.RegisterMessage(MLWorkerInfo.InstalledPackagesEntry)

PlatformInfo = _reflection.GeneratedProtocolMessageType('PlatformInfo', (_message.Message,), {
  'DESCRIPTOR' : _PLATFORMINFO,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.PlatformInfo)
  })
_sym_db.RegisterMessage(PlatformInfo)

FileUploadRequest = _reflection.GeneratedProtocolMessageType('FileUploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _FILEUPLOADREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.FileUploadRequest)
  })
_sym_db.RegisterMessage(FileUploadRequest)

FileUploadMetadata = _reflection.GeneratedProtocolMessageType('FileUploadMetadata', (_message.Message,), {
  'DESCRIPTOR' : _FILEUPLOADMETADATA,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.FileUploadMetadata)
  })
_sym_db.RegisterMessage(FileUploadMetadata)

EchoMsg = _reflection.GeneratedProtocolMessageType('EchoMsg', (_message.Message,), {
  'DESCRIPTOR' : _ECHOMSG,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.EchoMsg)
  })
_sym_db.RegisterMessage(EchoMsg)

ExplainRequest = _reflection.GeneratedProtocolMessageType('ExplainRequest', (_message.Message,), {

  'ColumnsEntry' : _reflection.GeneratedProtocolMessageType('ColumnsEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAINREQUEST_COLUMNSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainRequest.ColumnsEntry)
    })
  ,
  'DESCRIPTOR' : _EXPLAINREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ExplainRequest)
  })
_sym_db.RegisterMessage(ExplainRequest)
_sym_db.RegisterMessage(ExplainRequest.ColumnsEntry)

ExplainTextRequest = _reflection.GeneratedProtocolMessageType('ExplainTextRequest', (_message.Message,), {

  'ColumnsEntry' : _reflection.GeneratedProtocolMessageType('ColumnsEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAINTEXTREQUEST_COLUMNSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainTextRequest.ColumnsEntry)
    })
  ,

  'ColumnTypesEntry' : _reflection.GeneratedProtocolMessageType('ColumnTypesEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAINTEXTREQUEST_COLUMNTYPESENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainTextRequest.ColumnTypesEntry)
    })
  ,
  'DESCRIPTOR' : _EXPLAINTEXTREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ExplainTextRequest)
  })
_sym_db.RegisterMessage(ExplainTextRequest)
_sym_db.RegisterMessage(ExplainTextRequest.ColumnsEntry)
_sym_db.RegisterMessage(ExplainTextRequest.ColumnTypesEntry)

ExplainResponse = _reflection.GeneratedProtocolMessageType('ExplainResponse', (_message.Message,), {

  'Explanation' : _reflection.GeneratedProtocolMessageType('Explanation', (_message.Message,), {

    'PerFeatureEntry' : _reflection.GeneratedProtocolMessageType('PerFeatureEntry', (_message.Message,), {
      'DESCRIPTOR' : _EXPLAINRESPONSE_EXPLANATION_PERFEATUREENTRY,
      '__module__' : 'ml_worker_pb2'
      # @@protoc_insertion_point(class_scope:worker.ExplainResponse.Explanation.PerFeatureEntry)
      })
    ,
    'DESCRIPTOR' : _EXPLAINRESPONSE_EXPLANATION,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainResponse.Explanation)
    })
  ,

  'ExplanationsEntry' : _reflection.GeneratedProtocolMessageType('ExplanationsEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAINRESPONSE_EXPLANATIONSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainResponse.ExplanationsEntry)
    })
  ,
  'DESCRIPTOR' : _EXPLAINRESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ExplainResponse)
  })
_sym_db.RegisterMessage(ExplainResponse)
_sym_db.RegisterMessage(ExplainResponse.Explanation)
_sym_db.RegisterMessage(ExplainResponse.Explanation.PerFeatureEntry)
_sym_db.RegisterMessage(ExplainResponse.ExplanationsEntry)

ExplainTextResponse = _reflection.GeneratedProtocolMessageType('ExplainTextResponse', (_message.Message,), {

  'WeightsPerFeature' : _reflection.GeneratedProtocolMessageType('WeightsPerFeature', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAINTEXTRESPONSE_WEIGHTSPERFEATURE,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainTextResponse.WeightsPerFeature)
    })
  ,

  'WeightsEntry' : _reflection.GeneratedProtocolMessageType('WeightsEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAINTEXTRESPONSE_WEIGHTSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.ExplainTextResponse.WeightsEntry)
    })
  ,
  'DESCRIPTOR' : _EXPLAINTEXTRESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ExplainTextResponse)
  })
_sym_db.RegisterMessage(ExplainTextResponse)
_sym_db.RegisterMessage(ExplainTextResponse.WeightsPerFeature)
_sym_db.RegisterMessage(ExplainTextResponse.WeightsEntry)

RunModelForDataFrameResponse = _reflection.GeneratedProtocolMessageType('RunModelForDataFrameResponse', (_message.Message,), {
  'DESCRIPTOR' : _RUNMODELFORDATAFRAMERESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunModelForDataFrameResponse)
  })
_sym_db.RegisterMessage(RunModelForDataFrameResponse)

DataRow = _reflection.GeneratedProtocolMessageType('DataRow', (_message.Message,), {

  'ColumnsEntry' : _reflection.GeneratedProtocolMessageType('ColumnsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATAROW_COLUMNSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.DataRow.ColumnsEntry)
    })
  ,
  'DESCRIPTOR' : _DATAROW,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DataRow)
  })
_sym_db.RegisterMessage(DataRow)
_sym_db.RegisterMessage(DataRow.ColumnsEntry)

DataFrame = _reflection.GeneratedProtocolMessageType('DataFrame', (_message.Message,), {
  'DESCRIPTOR' : _DATAFRAME,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DataFrame)
  })
_sym_db.RegisterMessage(DataFrame)

RunModelForDataFrameRequest = _reflection.GeneratedProtocolMessageType('RunModelForDataFrameRequest', (_message.Message,), {

  'ColumnTypesEntry' : _reflection.GeneratedProtocolMessageType('ColumnTypesEntry', (_message.Message,), {
    'DESCRIPTOR' : _RUNMODELFORDATAFRAMEREQUEST_COLUMNTYPESENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.RunModelForDataFrameRequest.ColumnTypesEntry)
    })
  ,

  'ColumnDtypesEntry' : _reflection.GeneratedProtocolMessageType('ColumnDtypesEntry', (_message.Message,), {
    'DESCRIPTOR' : _RUNMODELFORDATAFRAMEREQUEST_COLUMNDTYPESENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.RunModelForDataFrameRequest.ColumnDtypesEntry)
    })
  ,
  'DESCRIPTOR' : _RUNMODELFORDATAFRAMEREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunModelForDataFrameRequest)
  })
_sym_db.RegisterMessage(RunModelForDataFrameRequest)
_sym_db.RegisterMessage(RunModelForDataFrameRequest.ColumnTypesEntry)
_sym_db.RegisterMessage(RunModelForDataFrameRequest.ColumnDtypesEntry)

RunModelRequest = _reflection.GeneratedProtocolMessageType('RunModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _RUNMODELREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunModelRequest)
  })
_sym_db.RegisterMessage(RunModelRequest)

RunModelResponse = _reflection.GeneratedProtocolMessageType('RunModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _RUNMODELRESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunModelResponse)
  })
_sym_db.RegisterMessage(RunModelResponse)

RunAdHocTestRequest = _reflection.GeneratedProtocolMessageType('RunAdHocTestRequest', (_message.Message,), {
  'DESCRIPTOR' : _RUNADHOCTESTREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunAdHocTestRequest)
  })
_sym_db.RegisterMessage(RunAdHocTestRequest)

DatasetProcessingRequest = _reflection.GeneratedProtocolMessageType('DatasetProcessingRequest', (_message.Message,), {
  'DESCRIPTOR' : _DATASETPROCESSINGREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DatasetProcessingRequest)
  })
_sym_db.RegisterMessage(DatasetProcessingRequest)

DatasetProcessingFunction = _reflection.GeneratedProtocolMessageType('DatasetProcessingFunction', (_message.Message,), {
  'DESCRIPTOR' : _DATASETPROCESSINGFUNCTION,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DatasetProcessingFunction)
  })
_sym_db.RegisterMessage(DatasetProcessingFunction)

SuiteTestArgument = _reflection.GeneratedProtocolMessageType('SuiteTestArgument', (_message.Message,), {
  'DESCRIPTOR' : _SUITETESTARGUMENT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.SuiteTestArgument)
  })
_sym_db.RegisterMessage(SuiteTestArgument)

RunTestSuiteRequest = _reflection.GeneratedProtocolMessageType('RunTestSuiteRequest', (_message.Message,), {
  'DESCRIPTOR' : _RUNTESTSUITEREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunTestSuiteRequest)
  })
_sym_db.RegisterMessage(RunTestSuiteRequest)

FuncArgument = _reflection.GeneratedProtocolMessageType('FuncArgument', (_message.Message,), {
  'DESCRIPTOR' : _FUNCARGUMENT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.FuncArgument)
  })
_sym_db.RegisterMessage(FuncArgument)

RunTestResponse = _reflection.GeneratedProtocolMessageType('RunTestResponse', (_message.Message,), {
  'DESCRIPTOR' : _RUNTESTRESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.RunTestResponse)
  })
_sym_db.RegisterMessage(RunTestResponse)

Partial_unexpected_counts = _reflection.GeneratedProtocolMessageType('Partial_unexpected_counts', (_message.Message,), {
  'DESCRIPTOR' : _PARTIAL_UNEXPECTED_COUNTS,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.Partial_unexpected_counts)
  })
_sym_db.RegisterMessage(Partial_unexpected_counts)

NamedSingleTestResult = _reflection.GeneratedProtocolMessageType('NamedSingleTestResult', (_message.Message,), {
  'DESCRIPTOR' : _NAMEDSINGLETESTRESULT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.NamedSingleTestResult)
  })
_sym_db.RegisterMessage(NamedSingleTestResult)

IdentifierSingleTestResult = _reflection.GeneratedProtocolMessageType('IdentifierSingleTestResult', (_message.Message,), {
  'DESCRIPTOR' : _IDENTIFIERSINGLETESTRESULT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.IdentifierSingleTestResult)
  })
_sym_db.RegisterMessage(IdentifierSingleTestResult)

TestMessage = _reflection.GeneratedProtocolMessageType('TestMessage', (_message.Message,), {
  'DESCRIPTOR' : _TESTMESSAGE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.TestMessage)
  })
_sym_db.RegisterMessage(TestMessage)

SingleTestResult = _reflection.GeneratedProtocolMessageType('SingleTestResult', (_message.Message,), {

  'PropsEntry' : _reflection.GeneratedProtocolMessageType('PropsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SINGLETESTRESULT_PROPSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.SingleTestResult.PropsEntry)
    })
  ,
  'DESCRIPTOR' : _SINGLETESTRESULT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.SingleTestResult)
  })
_sym_db.RegisterMessage(SingleTestResult)
_sym_db.RegisterMessage(SingleTestResult.PropsEntry)

TestResultMessage = _reflection.GeneratedProtocolMessageType('TestResultMessage', (_message.Message,), {
  'DESCRIPTOR' : _TESTRESULTMESSAGE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.TestResultMessage)
  })
_sym_db.RegisterMessage(TestResultMessage)

DatasetProcessingResultMessage = _reflection.GeneratedProtocolMessageType('DatasetProcessingResultMessage', (_message.Message,), {
  'DESCRIPTOR' : _DATASETPROCESSINGRESULTMESSAGE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DatasetProcessingResultMessage)
  })
_sym_db.RegisterMessage(DatasetProcessingResultMessage)

DatasetRowModificationResult = _reflection.GeneratedProtocolMessageType('DatasetRowModificationResult', (_message.Message,), {

  'ModificationsEntry' : _reflection.GeneratedProtocolMessageType('ModificationsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETROWMODIFICATIONRESULT_MODIFICATIONSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.DatasetRowModificationResult.ModificationsEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETROWMODIFICATIONRESULT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DatasetRowModificationResult)
  })
_sym_db.RegisterMessage(DatasetRowModificationResult)
_sym_db.RegisterMessage(DatasetRowModificationResult.ModificationsEntry)

TestSuiteResultMessage = _reflection.GeneratedProtocolMessageType('TestSuiteResultMessage', (_message.Message,), {
  'DESCRIPTOR' : _TESTSUITERESULTMESSAGE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.TestSuiteResultMessage)
  })
_sym_db.RegisterMessage(TestSuiteResultMessage)

ArtifactRef = _reflection.GeneratedProtocolMessageType('ArtifactRef', (_message.Message,), {
  'DESCRIPTOR' : _ARTIFACTREF,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ArtifactRef)
  })
_sym_db.RegisterMessage(ArtifactRef)

Chunk = _reflection.GeneratedProtocolMessageType('Chunk', (_message.Message,), {
  'DESCRIPTOR' : _CHUNK,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.Chunk)
  })
_sym_db.RegisterMessage(Chunk)

UploadStatus = _reflection.GeneratedProtocolMessageType('UploadStatus', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADSTATUS,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.UploadStatus)
  })
_sym_db.RegisterMessage(UploadStatus)

MLWorkerErrorInfo = _reflection.GeneratedProtocolMessageType('MLWorkerErrorInfo', (_message.Message,), {
  'DESCRIPTOR' : _MLWORKERERRORINFO,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.MLWorkerErrorInfo)
  })
_sym_db.RegisterMessage(MLWorkerErrorInfo)

SuiteInput = _reflection.GeneratedProtocolMessageType('SuiteInput', (_message.Message,), {
  'DESCRIPTOR' : _SUITEINPUT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.SuiteInput)
  })
_sym_db.RegisterMessage(SuiteInput)

ModelMeta = _reflection.GeneratedProtocolMessageType('ModelMeta', (_message.Message,), {
  'DESCRIPTOR' : _MODELMETA,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ModelMeta)
  })
_sym_db.RegisterMessage(ModelMeta)

DatasetMeta = _reflection.GeneratedProtocolMessageType('DatasetMeta', (_message.Message,), {
  'DESCRIPTOR' : _DATASETMETA,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DatasetMeta)
  })
_sym_db.RegisterMessage(DatasetMeta)

GenerateTestSuiteRequest = _reflection.GeneratedProtocolMessageType('GenerateTestSuiteRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATETESTSUITEREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.GenerateTestSuiteRequest)
  })
_sym_db.RegisterMessage(GenerateTestSuiteRequest)

GenerateTestSuiteResponse = _reflection.GeneratedProtocolMessageType('GenerateTestSuiteResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATETESTSUITERESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.GenerateTestSuiteResponse)
  })
_sym_db.RegisterMessage(GenerateTestSuiteResponse)

GeneratedTest = _reflection.GeneratedProtocolMessageType('GeneratedTest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEDTEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.GeneratedTest)
  })
_sym_db.RegisterMessage(GeneratedTest)

GeneratedTestInput = _reflection.GeneratedProtocolMessageType('GeneratedTestInput', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEDTESTINPUT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.GeneratedTestInput)
  })
_sym_db.RegisterMessage(GeneratedTestInput)

CatalogResponse = _reflection.GeneratedProtocolMessageType('CatalogResponse', (_message.Message,), {

  'TestsEntry' : _reflection.GeneratedProtocolMessageType('TestsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CATALOGRESPONSE_TESTSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.CatalogResponse.TestsEntry)
    })
  ,

  'SlicesEntry' : _reflection.GeneratedProtocolMessageType('SlicesEntry', (_message.Message,), {
    'DESCRIPTOR' : _CATALOGRESPONSE_SLICESENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.CatalogResponse.SlicesEntry)
    })
  ,

  'TransformationsEntry' : _reflection.GeneratedProtocolMessageType('TransformationsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CATALOGRESPONSE_TRANSFORMATIONSENTRY,
    '__module__' : 'ml_worker_pb2'
    # @@protoc_insertion_point(class_scope:worker.CatalogResponse.TransformationsEntry)
    })
  ,
  'DESCRIPTOR' : _CATALOGRESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.CatalogResponse)
  })
_sym_db.RegisterMessage(CatalogResponse)
_sym_db.RegisterMessage(CatalogResponse.TestsEntry)
_sym_db.RegisterMessage(CatalogResponse.SlicesEntry)
_sym_db.RegisterMessage(CatalogResponse.TransformationsEntry)

FunctionMeta = _reflection.GeneratedProtocolMessageType('FunctionMeta', (_message.Message,), {
  'DESCRIPTOR' : _FUNCTIONMETA,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.FunctionMeta)
  })
_sym_db.RegisterMessage(FunctionMeta)

DatasetProcessFunctionMeta = _reflection.GeneratedProtocolMessageType('DatasetProcessFunctionMeta', (_message.Message,), {
  'DESCRIPTOR' : _DATASETPROCESSFUNCTIONMETA,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.DatasetProcessFunctionMeta)
  })
_sym_db.RegisterMessage(DatasetProcessFunctionMeta)

TestFunctionArgument = _reflection.GeneratedProtocolMessageType('TestFunctionArgument', (_message.Message,), {
  'DESCRIPTOR' : _TESTFUNCTIONARGUMENT,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.TestFunctionArgument)
  })
_sym_db.RegisterMessage(TestFunctionArgument)

ComparisonClause = _reflection.GeneratedProtocolMessageType('ComparisonClause', (_message.Message,), {
  'DESCRIPTOR' : _COMPARISONCLAUSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.ComparisonClause)
  })
_sym_db.RegisterMessage(ComparisonClause)

GenerateQueryBasedSliceRequest = _reflection.GeneratedProtocolMessageType('GenerateQueryBasedSliceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEQUERYBASEDSLICEREQUEST,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.GenerateQueryBasedSliceRequest)
  })
_sym_db.RegisterMessage(GenerateQueryBasedSliceRequest)

GenerateQueryBasedSliceResponse = _reflection.GeneratedProtocolMessageType('GenerateQueryBasedSliceResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEQUERYBASEDSLICERESPONSE,
  '__module__' : 'ml_worker_pb2'
  # @@protoc_insertion_point(class_scope:worker.GenerateQueryBasedSliceResponse)
  })
_sym_db.RegisterMessage(GenerateQueryBasedSliceResponse)

_MLWORKER = DESCRIPTOR.services_by_name['MLWorker']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021ai.giskard.workerB\013WorkerProtoP\001\242\002\003WRK'
  _MLWORKERINFO_INSTALLEDPACKAGESENTRY._options = None
  _MLWORKERINFO_INSTALLEDPACKAGESENTRY._serialized_options = b'8\001'
  _EXPLAINREQUEST_COLUMNSENTRY._options = None
  _EXPLAINREQUEST_COLUMNSENTRY._serialized_options = b'8\001'
  _EXPLAINTEXTREQUEST_COLUMNSENTRY._options = None
  _EXPLAINTEXTREQUEST_COLUMNSENTRY._serialized_options = b'8\001'
  _EXPLAINTEXTREQUEST_COLUMNTYPESENTRY._options = None
  _EXPLAINTEXTREQUEST_COLUMNTYPESENTRY._serialized_options = b'8\001'
  _EXPLAINRESPONSE_EXPLANATION_PERFEATUREENTRY._options = None
  _EXPLAINRESPONSE_EXPLANATION_PERFEATUREENTRY._serialized_options = b'8\001'
  _EXPLAINRESPONSE_EXPLANATIONSENTRY._options = None
  _EXPLAINRESPONSE_EXPLANATIONSENTRY._serialized_options = b'8\001'
  _EXPLAINTEXTRESPONSE_WEIGHTSENTRY._options = None
  _EXPLAINTEXTRESPONSE_WEIGHTSENTRY._serialized_options = b'8\001'
  _DATAROW_COLUMNSENTRY._options = None
  _DATAROW_COLUMNSENTRY._serialized_options = b'8\001'
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNTYPESENTRY._options = None
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNTYPESENTRY._serialized_options = b'8\001'
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNDTYPESENTRY._options = None
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNDTYPESENTRY._serialized_options = b'8\001'
  _SINGLETESTRESULT_PROPSENTRY._options = None
  _SINGLETESTRESULT_PROPSENTRY._serialized_options = b'8\001'
  _DATASETROWMODIFICATIONRESULT_MODIFICATIONSENTRY._options = None
  _DATASETROWMODIFICATIONRESULT_MODIFICATIONSENTRY._serialized_options = b'8\001'
  _CATALOGRESPONSE_TESTSENTRY._options = None
  _CATALOGRESPONSE_TESTSENTRY._serialized_options = b'8\001'
  _CATALOGRESPONSE_SLICESENTRY._options = None
  _CATALOGRESPONSE_SLICESENTRY._serialized_options = b'8\001'
  _CATALOGRESPONSE_TRANSFORMATIONSENTRY._options = None
  _CATALOGRESPONSE_TRANSFORMATIONSENTRY._serialized_options = b'8\001'
  _FILETYPE._serialized_start=7394
  _FILETYPE._serialized_end=7428
  _TESTMESSAGETYPE._serialized_start=7430
  _TESTMESSAGETYPE._serialized_end=7468
  _STATUSCODE._serialized_start=7470
  _STATUSCODE._serialized_end=7551
  _COMPARISONTYPE._serialized_start=7554
  _COMPARISONTYPE._serialized_end=7691
  _MLWORKERINFOREQUEST._serialized_start=115
  _MLWORKERINFOREQUEST._serialized_end=159
  _MLWORKERINFO._serialized_start=162
  _MLWORKERINFO._serialized_end=520
  _MLWORKERINFO_INSTALLEDPACKAGESENTRY._serialized_start=464
  _MLWORKERINFO_INSTALLEDPACKAGESENTRY._serialized_end=520
  _PLATFORMINFO._serialized_start=522
  _PLATFORMINFO._serialized_end=636
  _FILEUPLOADREQUEST._serialized_start=638
  _FILEUPLOADREQUEST._serialized_end=748
  _FILEUPLOADMETADATA._serialized_start=750
  _FILEUPLOADMETADATA._serialized_end=854
  _ECHOMSG._serialized_start=856
  _ECHOMSG._serialized_end=878
  _EXPLAINREQUEST._serialized_start=881
  _EXPLAINREQUEST._serialized_end=1073
  _EXPLAINREQUEST_COLUMNSENTRY._serialized_start=1027
  _EXPLAINREQUEST_COLUMNSENTRY._serialized_end=1073
  _EXPLAINTEXTREQUEST._serialized_start=1076
  _EXPLAINTEXTREQUEST._serialized_end=1398
  _EXPLAINTEXTREQUEST_COLUMNSENTRY._serialized_start=1027
  _EXPLAINTEXTREQUEST_COLUMNSENTRY._serialized_end=1073
  _EXPLAINTEXTREQUEST_COLUMNTYPESENTRY._serialized_start=1348
  _EXPLAINTEXTREQUEST_COLUMNTYPESENTRY._serialized_end=1398
  _EXPLAINRESPONSE._serialized_start=1401
  _EXPLAINRESPONSE._serialized_end=1714
  _EXPLAINRESPONSE_EXPLANATION._serialized_start=1486
  _EXPLAINRESPONSE_EXPLANATION._serialized_end=1624
  _EXPLAINRESPONSE_EXPLANATION_PERFEATUREENTRY._serialized_start=1575
  _EXPLAINRESPONSE_EXPLANATION_PERFEATUREENTRY._serialized_end=1624
  _EXPLAINRESPONSE_EXPLANATIONSENTRY._serialized_start=1626
  _EXPLAINRESPONSE_EXPLANATIONSENTRY._serialized_end=1714
  _EXPLAINTEXTRESPONSE._serialized_start=1717
  _EXPLAINTEXTRESPONSE._serialized_end=1945
  _EXPLAINTEXTRESPONSE_WEIGHTSPERFEATURE._serialized_start=1814
  _EXPLAINTEXTRESPONSE_WEIGHTSPERFEATURE._serialized_end=1850
  _EXPLAINTEXTRESPONSE_WEIGHTSENTRY._serialized_start=1852
  _EXPLAINTEXTRESPONSE_WEIGHTSENTRY._serialized_end=1945
  _RUNMODELFORDATAFRAMERESPONSE._serialized_start=1948
  _RUNMODELFORDATAFRAMERESPONSE._serialized_end=2089
  _DATAROW._serialized_start=2091
  _DATAROW._serialized_end=2195
  _DATAROW_COLUMNSENTRY._serialized_start=1027
  _DATAROW_COLUMNSENTRY._serialized_end=1073
  _DATAFRAME._serialized_start=2197
  _DATAFRAME._serialized_end=2239
  _RUNMODELFORDATAFRAMEREQUEST._serialized_start=2242
  _RUNMODELFORDATAFRAMEREQUEST._serialized_end=2620
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNTYPESENTRY._serialized_start=1348
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNTYPESENTRY._serialized_end=1398
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNDTYPESENTRY._serialized_start=2569
  _RUNMODELFORDATAFRAMEREQUEST_COLUMNDTYPESENTRY._serialized_end=2620
  _RUNMODELREQUEST._serialized_start=2623
  _RUNMODELREQUEST._serialized_end=2757
  _RUNMODELRESPONSE._serialized_start=2759
  _RUNMODELRESPONSE._serialized_end=2822
  _RUNADHOCTESTREQUEST._serialized_start=2824
  _RUNADHOCTESTREQUEST._serialized_end=2904
  _DATASETPROCESSINGREQUEST._serialized_start=2906
  _DATASETPROCESSINGREQUEST._serialized_end=3024
  _DATASETPROCESSINGFUNCTION._serialized_start=3027
  _DATASETPROCESSINGFUNCTION._serialized_end=3210
  _SUITETESTARGUMENT._serialized_start=3212
  _SUITETESTARGUMENT._serialized_end=3302
  _RUNTESTSUITEREQUEST._serialized_start=3304
  _RUNTESTSUITEREQUEST._serialized_end=3414
  _FUNCARGUMENT._serialized_start=3417
  _FUNCARGUMENT._serialized_end=3769
  _RUNTESTRESPONSE._serialized_start=3771
  _RUNTESTRESPONSE._serialized_end=3802
  _PARTIAL_UNEXPECTED_COUNTS._serialized_start=3804
  _PARTIAL_UNEXPECTED_COUNTS._serialized_end=3861
  _NAMEDSINGLETESTRESULT._serialized_start=3863
  _NAMEDSINGLETESTRESULT._serialized_end=3946
  _IDENTIFIERSINGLETESTRESULT._serialized_start=3948
  _IDENTIFIERSINGLETESTRESULT._serialized_end=4030
  _TESTMESSAGE._serialized_start=4032
  _TESTMESSAGE._serialized_end=4098
  _SINGLETESTRESULT._serialized_start=4101
  _SINGLETESTRESULT._serialized_end=4733
  _SINGLETESTRESULT_PROPSENTRY._serialized_start=4689
  _SINGLETESTRESULT_PROPSENTRY._serialized_end=4733
  _TESTRESULTMESSAGE._serialized_start=4735
  _TESTRESULTMESSAGE._serialized_end=4802
  _DATASETPROCESSINGRESULTMESSAGE._serialized_start=4805
  _DATASETPROCESSINGRESULTMESSAGE._serialized_end=4958
  _DATASETROWMODIFICATIONRESULT._serialized_start=4961
  _DATASETROWMODIFICATIONRESULT._serialized_end=5140
  _DATASETROWMODIFICATIONRESULT_MODIFICATIONSENTRY._serialized_start=5088
  _DATASETROWMODIFICATIONRESULT_MODIFICATIONSENTRY._serialized_end=5140
  _TESTSUITERESULTMESSAGE._serialized_start=5142
  _TESTSUITERESULTMESSAGE._serialized_end=5268
  _ARTIFACTREF._serialized_start=5270
  _ARTIFACTREF._serialized_end=5348
  _CHUNK._serialized_start=5350
  _CHUNK._serialized_end=5374
  _UPLOADSTATUS._serialized_start=5376
  _UPLOADSTATUS._serialized_end=5424
  _MLWORKERERRORINFO._serialized_start=5426
  _MLWORKERERRORINFO._serialized_end=5475
  _SUITEINPUT._serialized_start=5478
  _SUITEINPUT._serialized_end=5642
  _MODELMETA._serialized_start=5644
  _MODELMETA._serialized_end=5695
  _DATASETMETA._serialized_start=5697
  _DATASETMETA._serialized_end=5742
  _GENERATETESTSUITEREQUEST._serialized_start=5744
  _GENERATETESTSUITEREQUEST._serialized_end=5827
  _GENERATETESTSUITERESPONSE._serialized_start=5829
  _GENERATETESTSUITERESPONSE._serialized_end=5894
  _GENERATEDTEST._serialized_start=5896
  _GENERATEDTEST._serialized_end=5974
  _GENERATEDTESTINPUT._serialized_start=5976
  _GENERATEDTESTINPUT._serialized_end=6043
  _CATALOGRESPONSE._serialized_start=6046
  _CATALOGRESPONSE._serialized_end=6481
  _CATALOGRESPONSE_TESTSENTRY._serialized_start=6240
  _CATALOGRESPONSE_TESTSENTRY._serialized_end=6306
  _CATALOGRESPONSE_SLICESENTRY._serialized_start=6308
  _CATALOGRESPONSE_SLICESENTRY._serialized_end=6389
  _CATALOGRESPONSE_TRANSFORMATIONSENTRY._serialized_start=6391
  _CATALOGRESPONSE_TRANSFORMATIONSENTRY._serialized_end=6481
  _FUNCTIONMETA._serialized_start=6484
  _FUNCTIONMETA._serialized_end=6698
  _DATASETPROCESSFUNCTIONMETA._serialized_start=6701
  _DATASETPROCESSFUNCTIONMETA._serialized_end=7009
  _TESTFUNCTIONARGUMENT._serialized_start=7011
  _TESTFUNCTIONARGUMENT._serialized_end=7114
  _COMPARISONCLAUSE._serialized_start=7117
  _COMPARISONCLAUSE._serialized_end=7257
  _GENERATEQUERYBASEDSLICEREQUEST._serialized_start=7259
  _GENERATEQUERYBASEDSLICEREQUEST._serialized_end=7345
  _GENERATEQUERYBASEDSLICERESPONSE._serialized_start=7347
  _GENERATEQUERYBASEDSLICERESPONSE._serialized_end=7392
  _MLWORKER._serialized_start=7694
  _MLWORKER._serialized_end=8703
# @@protoc_insertion_point(module_scope)



import numpy as np
import pandas as pd
from dataclasses import dataclass

from giskard import Dataset
from giskard.models import cache as models_cache
from giskard.models.base import BaseModel
from giskard.scanner.decorators import detector
from giskard.scanner.issues import Issue
from giskard.scanner.logger import logger


@detector(name="stochasticity", tags=["stochasticity", "classification", "regression"])
class StochasticityDetector:
    def run(self, model: BaseModel, dataset: Dataset):
        logger.info("StochasticityDetector: Running")
        sample_size = min(100, len(dataset))
        reduced_dataset = dataset.slice(lambda df: df.sample(sample_size), row_level=False)

        with models_cache.no_cache():
            prediction_1 = model.predict(reduced_dataset).raw
            prediction_2 = model.predict(reduced_dataset).raw

        changed_samples = ~(np.isclose(prediction_1, prediction_2).all(axis=-1))

        if not changed_samples.any():
            return []

        fail_samples = pd.DataFrame(
            columns=["First prediction", "Second prediction"],
            index=reduced_dataset.df.loc[changed_samples].index,
        )
        fail_samples["First prediction"] = list(prediction_1[changed_samples])
        fail_samples["Second prediction"] = list(prediction_2[changed_samples])

        return [StochasticityIssue(model, dataset, level="medium", info=StochasticityInfo(samples=fail_samples))]


@dataclass
class StochasticityInfo:
    samples: pd.DataFrame


class StochasticityIssue(Issue):
    group = "Stochasticity"

    @property
    def domain(self) -> str:
        return ""

    @property
    def metric(self) -> str:
        return "Stochasticity"

    @property
    def deviation(self) -> str:
        return "Your model provides different results at each execution."

    @property
    def description(self) -> str:
        return f"We found {len(self.info.samples)} examples for which your model provides a different output at each execution."

    def examples(self, n=3) -> pd.DataFrame:
        return self.info.samples.head(n)

    @property
    def importance(self) -> float:
        return 1

    def generate_tests(self, with_names=False) -> list:
        return []



from dataclasses import dataclass
import pandas as pd
from sklearn.metrics import adjusted_mutual_info_score, mutual_info_score
from scipy import stats

from ..common.examples import ExampleExtractor
from ...ml_worker.testing.registry.slicing_function import SlicingFunction
from ..issues import Issue
from ...slicing.slice_finder import SliceFinder
from ..logger import logger
from ...datasets.base import Dataset
from ...models.base import BaseModel
from ..registry import Detector
from ..decorators import detector


@detector(name="spurious_correlation", tags=["spurious_correlation", "classification"])
class SpuriousCorrelationDetector(Detector):
    def __init__(self, method="theil", threshold=0.5) -> None:
        self.threshold = threshold
        self.method = method

    def run(self, model: BaseModel, dataset: Dataset):
        logger.info(f"{self.__class__.__name__}: Running")

        # Dataset prediction
        ds_predictions = pd.Series(model.predict(dataset).prediction, dataset.df.index)

        # Keep only interesting features
        features = model.meta.feature_names or dataset.columns.drop(dataset.target, errors="ignore")

        # Warm up text metadata
        for f in features:
            if dataset.column_types[f] == "text":
                dataset.column_meta[f, "text"]

        # Prepare dataset for slicing
        df = dataset.df.copy()
        if dataset.target is not None:
            df.drop(columns=dataset.target, inplace=True)
        df["__gsk__target"] = pd.Categorical(ds_predictions)
        wdata = Dataset(df, target="__gsk__target", column_types=dataset.column_types)
        wdata.load_metadata_from_instance(dataset.column_meta)

        # Find slices
        sliced_cols = SliceFinder("tree").run(wdata, features, target=wdata.target)

        measure_fn, measure_name = self._get_measure_fn()
        issues = []
        for col, slices in sliced_cols.items():
            if not slices:
                continue

            for slice_fn in slices:
                data_slice = dataset.slice(slice_fn)

                # Skip small slices
                if len(data_slice) < 20 or len(data_slice) < 0.05 * len(dataset):
                    continue

                dx = pd.DataFrame(
                    {
                        "feature": dataset.df.index.isin(data_slice.df.index).astype(int),
                        "prediction": ds_predictions,
                    },
                    index=dataset.df.index,
                )
                dx.dropna(inplace=True)

                metric_value = measure_fn(dx.feature, dx.prediction)
                logger.info(f"{self.__class__.__name__}: {slice_fn}\tAssociation = {metric_value:.3f}")

                if metric_value > self.threshold:
                    predictions = dx[dx.feature > 0].prediction.value_counts(normalize=True)
                    info = SpuriousCorrelationInfo(col, slice_fn, metric_value, measure_name, predictions)
                    issues.append(SpuriousCorrelationIssue(model, dataset, "info", info))

        return issues

    def _get_measure_fn(self):
        if self.method == "theil":
            return _theil_u, "Theil's U"
        if self.method == "mutual_information" or self.method == "mi":
            return _mutual_information, "Mutual information"
        if self.method == "cramer":
            return _cramer_v, "Cramer's V"
        raise ValueError(f"Unknown method `{self.method}`")


def _cramer_v(x, y):
    ct = pd.crosstab(x, y)
    return stats.contingency.association(ct, method="cramer")


def _mutual_information(x, y):
    return adjusted_mutual_info_score(x, y)


def _theil_u(x, y):
    return mutual_info_score(x, y) / stats.entropy(pd.Series(y).value_counts(normalize=True))


@dataclass
class SpuriousCorrelationInfo:
    feature: str
    slice_fn: SlicingFunction
    metric_value: float
    metric_name: str
    predictions: pd.DataFrame


class SpuriousCorrelationIssue(Issue):
    group = "Spurious correlation"

    @property
    def features(self):
        return [self.info.feature]

    @property
    def domain(self) -> str:
        return str(self.info.slice_fn)

    @property
    def metric(self) -> str:
        return f"Nominal association ({self.info.metric_name})"

    @property
    def deviation(self) -> str:
        plabel, p = self.info.predictions.index[0], self.info.predictions.iloc[0]

        return f"Prediction {self.dataset.target} = `{plabel}` for {p * 100:.2f}% of samples in the slice"

    @property
    def slicing_fn(self):
        return self.info.slice_fn

    @property
    def description(self) -> str:
        pred = self.model.predict(self.dataset.slice(self.info.slice_fn)).prediction
        classes = pd.Series(pred).value_counts(normalize=True)
        plabel, p = classes.index[0], classes.iloc[0]
        return f"Data slice {self.info.slice_fn} seems to be highly associated to prediction {self.dataset.target} = `{plabel}` ({p * 100:.2f}% of predictions in the data slice)."

    # @lru_cache
    def examples(self, n=3):
        extractor = ExampleExtractor(self)
        return extractor.get_examples_dataframe(n, with_prediction=1)

    @property
    def importance(self) -> float:
        return self.info.metric_value



import pandas as pd
import numpy as np

from .registry import MetadataProvider
import langdetect
from langdetect import DetectorFactory

DetectorFactory.seed = 0


class TextMetadataProvider(MetadataProvider):
    def __init__(self, name="text"):
        self.name = name

    def generate_metadata(self, values: pd.Series):
        # Ensure this is text encoded as a string
        values = values.astype(str)

        return pd.DataFrame(
            {
                "text_length": values.map(len),
                "avg_word_length": values.map(_avg_word_length),
                "charset": pd.Categorical(values.map(_detect_charset)),
                "avg_whitespace": values.map(_avg_whitespace),
                "avg_digits": values.map(_avg_digits),
                "language": values.map(_detect_lang),
            },
            index=values.index,
        )

    def supported_types(self):
        return ["text"]


def _detect_charset(text: str):
    import chardet

    charset = chardet.detect(text.encode("utf-8", errors="ignore"))["encoding"]
    return charset or "undefined"


def _avg_word_length(text: str):
    # @TODO: improve this
    words = text.split()
    if len(words) == 0:
        return 0.0
    return np.mean([len(w) for w in words])


def _avg_whitespace(text: str):
    chars = list(text)
    if len(chars) == 0:
        return 0.0
    return np.mean([c.isspace() for c in chars])


def _avg_digits(text: str):
    chars = list(text)
    if len(chars) == 0:
        return 0.0
    return np.mean([c.isdigit() for c in chars])


def _detect_lang(text: str):
    if len(text.split()) <= 5:
        return pd.NA
    try:
        detected = langdetect.detect_langs(text)
        language = detected[0].lang
    except langdetect.lang_detect_exception.LangDetectException:
        language = "unknown"
    return language



import re

import pytest

from giskard import Dataset
from giskard.models.sklearn import SKLearnModel
from tests.utils import MockedClient, match_model_id, match_url_patterns

model_name = "uploaded model"


def test_upload_df(diabetes_dataset: Dataset, diabetes_dataset_with_target: Dataset):
    artifact_url_pattern = re.compile(
        r"http://giskard-host:12345/api/v2/artifacts/test-project/datasets/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}/[data.csv.zst|giskard\-dataset\-meta.yaml]"
    )
    datasets_url_pattern = re.compile("http://giskard-host:12345/api/v2/project/test-project/datasets")

    with MockedClient() as (client, mr):
        saved_id = diabetes_dataset_with_target.upload(client, "test-project")
        match_model_id(saved_id)
        match_url_patterns(mr.request_history, artifact_url_pattern)
        match_url_patterns(mr.request_history, datasets_url_pattern)

        with pytest.raises(Exception) as e:
            Dataset(
                df=diabetes_dataset.df,
                column_types=diabetes_dataset.column_types,
                target=diabetes_dataset_with_target.target,
            )
        assert e.match(
            "Invalid target parameter: 'target' column is not present in the dataset "
            "with columns: \['age', 'sex', 'bmi', 'bp', 's1', 's2', 's3', 's4', 's5', 's6'\]"
        )  # noqa

        with pytest.raises(Exception) as e:
            diabetes_dataset.column_types = {"test": "test"}
            Dataset(
                df=diabetes_dataset.df,
                column_types=diabetes_dataset.column_types,
                target=diabetes_dataset_with_target.target,
            )
        assert e.match(
            "Invalid target parameter: 'target' column is not present in the dataset "
            "with columns: \['age', 'sex', 'bmi', 'bp', 's1', 's2', 's3', 's4', 's5', 's6'\]"
        )  # noqa


def _test_upload_model(model: SKLearnModel, ds: Dataset):
    artifact_url_pattern = re.compile(
        "http://giskard-host:12345/api/v2/artifacts/test-project/models/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}/.*"
    )
    models_url_pattern = re.compile("http://giskard-host:12345/api/v2/project/test-project/models")
    with MockedClient() as (client, mr):
        if model.is_regression:
            # Warning Scenario: classification_labels is sent for regression model
            with pytest.warns(UserWarning):
                model.upload(client, "test-project", ds)
        else:
            model.upload(client, "test-project", ds)

        match_model_id(model.id)
        match_url_patterns(mr.request_history, artifact_url_pattern)
        match_url_patterns(mr.request_history, models_url_pattern)


def _test_upload_model_exceptions(model: SKLearnModel, ds: Dataset):
    with MockedClient() as (client, mr):
        # Error Scenario : invalid feature_names
        with pytest.raises(Exception) as e:
            SKLearnModel(
                model=model.model,
                model_type=model.meta.model_type,
                feature_names=["some"],
                name=model_name,
                classification_labels=model.meta.classification_labels,
            ).upload(client, "test-project", ds)
        assert e.match("Value mentioned in feature_names is not available in validate_df")

        if model.is_classification:
            # Error Scenario: Target has values not declared in Classification Label
            with pytest.raises(Exception) as e:
                SKLearnModel(
                    model=model.model,
                    model_type=model.meta.model_type,
                    feature_names=model.meta.feature_names,
                    name=model_name,
                    classification_labels=[0, 1],
                ).upload(client, "test-project", ds)
            assert e.match(
                "Values .* in .* column are not declared in classification_labels parameter .* of the model: uploaded model"
            )  # noqa


@pytest.mark.parametrize(
    "data,model,",
    [
        ("german_credit_data", "german_credit_model"),
        ("diabetes_dataset", "linear_regression_diabetes"),
    ],
)
def test_upload_models(data, model, request):
    data = request.getfixturevalue(data)
    model = request.getfixturevalue(model)
    _test_upload_model(model, data)


@pytest.mark.parametrize(
    "data,model,", [("german_credit_data", "german_credit_model"), ("diabetes_dataset", "linear_regression_diabetes")]
)
def test_upload_models_exceptions(data, model, request):
    data = request.getfixturevalue(data)
    model = request.getfixturevalue(model)
    _test_upload_model_exceptions(model, data)



import asyncio
import collections
import functools
import hashlib
import logging
import os
import sys
from time import sleep

import click
from lockfile.pidlockfile import PIDLockFile
from pydantic import AnyHttpUrl, parse_obj_as

from giskard.path_utils import run_dir
from giskard.settings import settings

logger = logging.getLogger(__name__)
logging.getLogger().setLevel(logging.INFO)
logging.getLogger("giskard").setLevel(logging.INFO)


def common_options(func):
    @click.option(
        "--verbose",
        "-v",
        is_flag=True,
        callback=set_verbose,
        default=False,
        expose_value=False,
        is_eager=True,
        help="Enable verbose logging",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def set_verbose(_ctx, _param, value):
    if value:
        logging.getLogger("giskard").setLevel(logging.DEBUG)


def remove_stale_pid_file(pid_file):
    pid = pid_file.read_pid()
    if pid is not None and is_pid_stale(pid):
        logger.debug("Stale PID file found, removing it")
        pid_file.break_lock()


def is_pid_stale(pid):
    try:
        os.kill(pid, 0)  # NOSONAR
    except (OSError, TypeError):
        return True
    else:
        return False


def create_pid_file_path(is_server, url):
    hash_value = ml_worker_id(is_server, url)
    return run_dir / f"ml-worker-{hash_value}.pid"


def ml_worker_id(is_server, url):
    key = f"{sys.executable}"
    if not is_server:
        key += url
    hash_value = hashlib.sha1(key.encode()).hexdigest()
    return hash_value


def validate_url(_ctx, _param, value) -> AnyHttpUrl:
    return parse_obj_as(AnyHttpUrl, value)


def run_daemon(is_server, url, api_key):
    from giskard.ml_worker.ml_worker import MLWorker
    from daemon import DaemonContext
    from daemon.daemon import change_working_directory

    log_path = get_log_path()
    logger.info(f"Writing logs to {log_path}")
    pid_file = PIDLockFile(create_pid_file_path(is_server, url))

    with DaemonContext(pidfile=pid_file, stdout=open(log_path, "w+t")):
        # For some reason requests.utils.should_bypass_proxies that's called inside every request made by requests
        # hangs when the process runs as a daemon. A dirty temporary fix is to disable proxies for daemon mode.
        # True reasons for this to happen to be investigated
        os.environ["no_proxy"] = "*"

        workdir = settings.home_dir / "tmp" / f"daemon-run-{os.getpid()}"
        workdir.mkdir(exist_ok=True, parents=True)
        change_working_directory(workdir)

        logger.info(f"Daemon PID: {os.getpid()}")
        asyncio.get_event_loop().run_until_complete(MLWorker(is_server, url, api_key).start())


def get_log_path():
    return run_dir / "ml-worker.log"


def tail(filename, n=100):
    """Return the last n lines of a file"""
    return collections.deque(open(filename), n)


def follow_file(filename):
    if not os.path.exists(filename):
        print(f"{filename} does not exists")
        return

    wait = 1
    with open(filename) as fp:
        exit_pooling = False
        skip_existing = True
        while not exit_pooling:
            line = fp.readline()

            if not line:
                skip_existing = False
                sleep(wait)
            elif not skip_existing:
                print(line, end="")



import asyncio
import functools
import logging
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Iterable, Union, Awaitable

import grpc
from google.protobuf import any_pb2
from google.protobuf.message import Message
from google.rpc.status_pb2 import Status
from grpc import ServicerContext, StatusCode, aio
from grpc.experimental import wrap_server_method_handler
from grpc_status import rpc_status

from giskard.ml_worker.exceptions.IllegalArgumentError import CodedError
from giskard.ml_worker.generated.ml_worker_pb2 import MLWorkerErrorInfo

MESSAGE_TYPE = Union[Message, Iterable[Message]]

logger = logging.getLogger(__name__)

pool = ThreadPoolExecutor(thread_name_prefix="ml_worker_thread")


class MLWorkerRequestInterceptor(grpc.aio.ServerInterceptor):
    @staticmethod
    async def terminate_with_exception(error_code: StatusCode, e: Exception, context: ServicerContext):
        detail = any_pb2.Any()
        detail.Pack(
            MLWorkerErrorInfo(
                stack=traceback.format_exc(),
                error=str(e),
            )
        )
        code, _ = error_code.value
        rich_status = Status(code=code, message=e.__class__.__name__, details=[detail])
        await context.abort_with_status(rpc_status.to_status(rich_status))

    async def intercept_service(
        self,
        continuation: Callable[[grpc.HandlerCallDetails], Awaitable[grpc.RpcMethodHandler]],
        handler_call_details: grpc.HandlerCallDetails,
    ) -> grpc.RpcMethodHandler:
        def _wrapper(behavior):
            @functools.wraps(behavior)
            async def wrapper(request, context: aio.ServicerContext):
                try:
                    loop = asyncio.get_running_loop()
                    res = await loop.run_in_executor(pool, behavior, request, context)
                    return res
                except CodedError as e:
                    logger.exception(e)
                    await MLWorkerRequestInterceptor.terminate_with_exception(e.code, e, context)
                except Exception as e:
                    logger.exception(e)
                    await MLWorkerRequestInterceptor.terminate_with_exception(StatusCode.INTERNAL, e, context)

            return wrapper

        handler = await continuation(handler_call_details)
        if handler and (handler.request_streaming or handler.response_streaming):
            return handler

        return wrap_server_method_handler(_wrapper, handler)



import datetime
import uuid
import warnings
from collections import Counter
from time import perf_counter
from typing import Optional, Sequence


from .issues import Issue
from .logger import logger
from .registry import DetectorRegistry
from .result import ScanResult
from ..core.model_validation import validate_model
from ..datasets.base import Dataset
from ..models.base import BaseModel
from ..utils import fullname
from ..utils.analytics_collector import analytics, analytics_method, get_dataset_properties, get_model_properties
from giskard.client.python_utils import warning

MAX_ISSUES_PER_DETECTOR = 15


class Scanner:
    def __init__(self, params: Optional[dict] = None, only=None):
        if isinstance(only, str):
            only = [only]

        self.params = params or dict()
        self.only = only
        self.uuid = uuid.uuid4()

    def analyze(
        self, model: BaseModel, dataset: Optional[Dataset] = None, verbose=True, raise_exceptions=False
    ) -> ScanResult:
        """Runs the analysis of a model and dataset, detecting issues."""
        if model.is_text_generation:
            logger.warning(
                "LLM support is in alpha version — 🔥 things may break ! Please report any issues to https://github.com/giskard-AI/giskard/issues."
            )

        model, dataset = self._prepare_model_dataset(model, dataset)

        if not model.is_text_generation:
            time_start = perf_counter()
            validate_model(model=model, validate_ds=dataset)
            model_validation_time = perf_counter() - time_start
        else:
            model_validation_time = None

        if not dataset.df.index.is_unique:
            warning(
                "You dataframe has duplicate indexes, which is currently not supported. "
                "We have to reset the dataframe index to avoid issues."
            )
            dataset = Dataset(
                df=dataset.df.reset_index(drop=True),
                name=dataset.name,
                target=dataset.target,
                column_types=dataset.column_types,
            )

        maybe_print("🔎 Running scan…", verbose=verbose)
        time_start = perf_counter()

        # Collect the detectors
        detectors = self.get_detectors(tags=[model.meta.model_type.value])

        # @TODO: this should be selective to specific warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            issues, errors = self._run_detectors(
                detectors, model, dataset, verbose=verbose, raise_exceptions=raise_exceptions
            )

        elapsed = perf_counter() - time_start
        maybe_print(
            f"Scan completed: {len(issues) or 'no'} issue{'s' if len(issues) != 1 else ''} found. (Took {datetime.timedelta(seconds=elapsed)})",
            verbose=verbose,
        )
        if errors:
            warning(
                f"{len(errors)} errors were encountered while running detectors. Please check the log to understand what went wrong. "
                "You can run the scan again with `raise_exceptions=True` to disable graceful handling."
            )

        issues = self._postprocess(issues)
        self._collect_analytics(model, dataset, issues, elapsed, model_validation_time)

        return ScanResult(issues)

    def _run_detectors(self, detectors, model, dataset, verbose=True, raise_exceptions=False):
        if not detectors:
            raise RuntimeError("No issue detectors available. Scan will not be performed.")

        logger.info(f"Running detectors: {[d.__class__.__name__ for d in detectors]}")

        issues = []
        errors = []
        for detector in detectors:
            maybe_print(f"Running detector {detector.__class__.__name__}…", end="", verbose=verbose)
            detector_start = perf_counter()
            try:
                detected_issues = detector.run(model, dataset)
            except Exception as err:
                logger.error(f"Detector {detector.__class__.__name__} failed with error: {err}")
                errors.append((detector, err))
                analytics.track(
                    "scan:run-detector:error",
                    {
                        "scan_uuid": self.uuid.hex,
                        "detector": fullname(detector),
                        "error": str(err),
                        "error_class": fullname(err),
                    },
                )
                if raise_exceptions:
                    raise err

                detected_issues = []
            detected_issues = sorted(detected_issues, key=lambda i: -i.importance)[:MAX_ISSUES_PER_DETECTOR]
            detector_elapsed = perf_counter() - detector_start
            maybe_print(
                f" {len(detected_issues)} issues detected. (Took {datetime.timedelta(seconds=detector_elapsed)})",
                verbose=verbose,
            )
            analytics.track(
                "scan:detector-run",
                {
                    "scan_uuid": self.uuid.hex,
                    "detector": fullname(detector),
                    "detector_elapsed": detector_elapsed,
                    "detected_issues": len(detected_issues),
                },
            )

            issues.extend(detected_issues)

        return issues, errors

    def _postprocess(self, issues: Sequence[Issue]) -> Sequence[Issue]:
        # If we detected a StochasticityIssue, we will have a possibly false
        # positive DataLeakageIssue. We remove it here.
        from .stochasticity.stochasticity_detector import StochasticityIssue
        from .data_leakage.data_leakage_detector import DataLeakageIssue

        if any(isinstance(issue, StochasticityIssue) for issue in issues):
            issues = [issue for issue in issues if not isinstance(issue, DataLeakageIssue)]

        return issues

    def _prepare_model_dataset(self, model: BaseModel, dataset: Optional[Dataset]):
        if model.is_text_generation and dataset is None:
            logger.warning(
                "No dataset provided. We will use TruthfulQA as a default dataset. This involves rewriting your model prompt."
            )
            from .llm.utils import load_default_dataset

            dataset = load_default_dataset()
            model = model.rewrite_prompt(
                template="{question}", input_variables=["question"], feature_names=["question"]
            )

            return model, dataset

        if dataset is None:
            raise ValueError(f"Dataset must be provided for {model.meta.model_type.value} models.")

        return model, dataset

    @analytics_method
    def _collect_analytics(self, model, dataset, issues, elapsed, model_validation_time):
        issues_counter = Counter([fullname(i) for i in issues]) if issues else {}

        properties = dict(
            elapsed=elapsed,
            model_validation_time=model_validation_time,
            total_issues=len(issues),
            **issues_counter,
        )
        properties.update(get_model_properties(model))
        properties.update(get_dataset_properties(dataset))

        analytics.track("scan", properties)

    def get_detectors(self, tags: Optional[Sequence[str]] = None) -> Sequence:
        """Returns the detector instances."""
        detectors = []
        classes = DetectorRegistry.get_detector_classes(tags=tags)

        # Filter detector classes
        if self.only:
            only_classes = DetectorRegistry.get_detector_classes(tags=self.only)
            keys_to_keep = set(only_classes.keys()).intersection(classes.keys())
            classes = {k: classes[k] for k in keys_to_keep}

        # Configure instances
        for name, detector_cls in classes.items():
            kwargs = self.params.get(name) or dict()
            detectors.append(detector_cls(**kwargs))

        return detectors


def maybe_print(*args, **kwargs):
    if kwargs.pop("verbose", True):
        print(*args, **kwargs)



import csv
import os
import uuid
from pathlib import Path
from typing import Dict, List, Any, Iterable, Optional

import numpy as np
import pandas as pd

from giskard.core.core import SupportedModelTypes
from giskard.settings import settings

NaN = float("NaN")

CACHE_CSV_FILENAME = "giskard-model-cache.csv"


def flatten(xs):
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


class ModelCache:
    id: Optional[str] = None
    prediction_cache: Dict[str, Any] = None

    vectorized_get_cache_or_na = None

    def __init__(self, model_type: SupportedModelTypes, id: Optional[str] = None, cache_dir: Path = None):
        self.id = id or str(uuid.uuid4())
        self.prediction_cache = dict()
        self.cache_dir = cache_dir or Path(settings.home_dir / settings.cache_dir / "global/prediction_cache" / self.id)

        if id is not None:
            if (self.cache_dir / CACHE_CSV_FILENAME).exists():
                with open(self.cache_dir / CACHE_CSV_FILENAME, "r") as pred_f:
                    reader = csv.reader(pred_f)
                    for row in reader:
                        if model_type == SupportedModelTypes.TEXT_GENERATION:
                            self.prediction_cache[row[0]] = row[1:]
                        elif model_type == SupportedModelTypes.REGRESSION:
                            self.prediction_cache[row[0]] = float(row[1])
                        else:
                            self.prediction_cache[row[0]] = [float(i) for i in row[1:]]

        self.vectorized_get_cache_or_na = np.vectorize(self.get_cache_or_na, otypes=[object])

    def get_cache_or_na(self, key: str):
        return self.prediction_cache.get(key, NaN)

    def read_from_cache(self, keys: pd.Series):
        return pd.Series(self.vectorized_get_cache_or_na(keys), index=keys.index)

    def set_cache(self, keys: pd.Series, values: List[Any]):
        for i in range(len(keys)):
            self.prediction_cache[keys.iloc[i]] = values[i]

        if self.id:
            os.makedirs(self.cache_dir, exist_ok=True)
            with open(self.cache_dir / CACHE_CSV_FILENAME, "a") as pred_f:
                writer = csv.writer(pred_f)
                for i in range(len(keys)):
                    writer.writerow(flatten([keys.iloc[i], values[i]]))

    def _to_df(self):
        index = [key for key, values in self.prediction_cache.items()]
        data = np.array([list(flatten([values])) for key, values in self.prediction_cache.items()])

        if len(data) > 0:
            return pd.DataFrame(data, index=index)
        else:
            return pd.DataFrame({})
